import slope
import slope.nn as nn
import math


class Net(nn.Module):
    def __init__(self, in_dim, embed_dim, out_dim):
        self.in_proj = nn.Linear(in_dim, embed_dim)
        self.bn_out = nn.BatchNorm(embed_dim)
        self.out_proj = nn.Linear(embed_dim, out_dim)

    def __call__(self, x, training=False):
        x = self.in_proj(x).relu()
        x = self.bn_out(x, training=training)
        x = self.out_proj(x)
        return (x, self) if training else x


def loss_fn(model, batch):
    x, y = batch
    y_hat, model = model(x, training=True)
    loss = (y - y_hat).pow(2).sum()
    return loss, model


vgloss_fn = slope.value_and_grad(loss_fn, has_aux=True)


@slope.jit
def train_step(model, batch, optimizer):
    ((loss, model), gloss_model) = vgloss_fn(model, batch)
    model, optimizer = optimizer(model, gloss_model)
    return loss, model, optimizer


x = slope.ones((3, 2))
y = slope.ones(3, 1)
batch = (x, y)
model = Net(2, 5, 1)
optim = nn.SGD(model, lr=1e-3, momentum=0.8, weight_decay=1e-5)
print(f"Before: {model.bn_out.running_mean=}")
loss, model, optimizer_ = train_step(model, batch, optim)
print(f"After: {model.bn_out.running_mean=}")



import slope
from slope.core import ProcedureSet, Tensor, dtypes
import math
import numpy as np
from typing import Tuple, List, Dict, Any, Optional, Sequence, Union, Iterator, NamedTuple, DefaultDict
from collections import defaultdict
import functools

max_ = max
abs_ = abs
min_ = min
sum_ = sum
procedure_set = ProcedureSet()


@procedure_set.register()
def zeros(*args, **kwargs):
    dtype = kwargs.get("dtype", slope.core.backend.DEFAULT_DTYPE)
    if kwargs.get("shape", None) is None:
        shape = args[0] if isinstance(args[0], (tuple, list)) else args
        assert all(i >= 0 for i in shape)
    return slope.full(shape, 0.0, dtype)


@procedure_set.register()
def ones(*args, **kwargs):
    dtype = kwargs.get("dtype", slope.core.backend.DEFAULT_DTYPE)
    if kwargs.get("shape", None) is None:
        shape = args[0] if isinstance(args[0], (tuple, list)) else args
        assert all(i >= 0 for i in shape)
    return slope.full(shape=shape, fill_value=1.0, dtype=dtype)


@procedure_set.register()
def full_like(y, fill_value):
    return slope.full(shape=y.shape, fill_value=fill_value, dtype=y.dtype)


@procedure_set.register()
def zeros_like(y):
    return full_like(y, 0.0)


@procedure_set.register()
def ones_like(y):
    return full_like(y, 1.0)


@procedure_set.register()
def rand_like(y):
    return slope.rand(shape=y.shape, dtype=y.dtype)


@procedure_set.register()
def randn_like(y):
    return slope.randn(shape=y.shape, dtype=y.dtype)


@procedure_set.register()
def eye(dim: int, **kwargs):
    return slope.ones((dim, 1)).pad((0, dim, 0, 0)).flatten().padslice(((0, dim * dim),)).reshape(dim, dim)


@procedure_set.register()
def where(x, trueval, falseval):
    if not isinstance(trueval, Tensor):
        trueval = slope.full((), trueval, device=x.device)
    if not isinstance(falseval, Tensor):
        falseval = slope.full((), falseval, device=x.device)
    cond = x != x.zeros_like()
    if not any(val.dtype is dtypes.bool for val in (trueval, falseval)):
        cond = cond.cast(trueval.dtype)
        return cond * trueval + (cond.ones_like() - cond) * falseval
    else:
        cond = cond.cast(slope.float32)
        trueval = trueval.cast(slope.float32)
        falseval = falseval.cast(slope.float32)
        return (cond * trueval + (cond.ones_like() - cond) * falseval).cast(dtypes.bool)


@procedure_set.register()
def mean(x, dim=None, keepdim=False):
    out = x.sum(dim=dim, keepdim=keepdim)
    return out * (math.prod(out.shape) / math.prod(x.shape))


@procedure_set.register()
def rsqrt(x):
    return (x.ones_like() / x).sqrt()


@procedure_set.register()
def cos(x):
    return (x.full_like(math.pi / 2) - x).sin()


@procedure_set.register()
def tan(x):
    return x.sin() / x.cos()


@procedure_set.register()
def neg(x):
    return x.full_like(-1) * x


@procedure_set.register()
def not_equal(x, w):
    return ~(x == w)


@procedure_set.register()
def greater_equal(x, w):
    return ~(x < w)


@procedure_set.register()
def less_equal(x, w):
    return ~(x > w)


@procedure_set.register()
def minimum(x, w):
    return -x.maximum(-x, -w)


@procedure_set.register()
def min(x, dim=None, keepdim=False):
    return -((-x).max(x, dim, keepdim))


@procedure_set.register()
def argmax(x, dim=None, keepdim=False):
    if dim is None:
        ar = slope.arange(math.prod(x.shape) - 1, -1, -1, dtype=x.dtype)
        ar = ar.reshape(x.shape)
        idx = (x == x.max(dim)).cast(x.dtype) * ar
        return math.prod(x.shape) - idx.max().cast(slope.int32) - 1
    dim = dim + len(x.shape) if dim < 0 else dim
    m = (x == x.max(dim=dim, keepdim=True)).cast(x.dtype)
    idx = m * slope.arange(x.shape[dim] - 1, -1, -1, dtype=x.dtype)
    # idx = idx.reshape(x.shape[dim], *[1] * (x.ndim - dim - 1))
    ret = x.shape[dim] - idx.max(dim=dim, keepdim=keepdim) - 1
    return ret.cast(slope.int32)


@procedure_set.register()
def argmin(x, dim=None, keepdim=False):
    return (-x).argmax(dim=dim, keepdim=keepdim)


@procedure_set.register()
def log2(x):
    return x.log() / math.log(2)


@procedure_set.register()
@staticmethod
def _tri(r: int, c: int, k: int = 0, **kwargs) -> Tensor:
    return slope.arange(r, **kwargs).unsqueeze(1).expand(r, c) <= slope.arange(-k, c - k, **kwargs).unsqueeze(0).expand(r, c)


@procedure_set.register()
def triu(x, k: int = 0) -> Tensor:
    return _tri(x.shape[-2], x.shape[-1], k=k, dtype=x.dtype, device=x.device).where(x, slope.zeros_like(x))


@procedure_set.register()
def tril(x, k: int = 0) -> Tensor:
    return _tri(x.shape[-2], x.shape[-1], k=k + 1, dtype=x.dtype, device=x.device).where(slope.zeros_like(x), x)


@procedure_set.register()
def trunc(x: Tensor) -> Tensor:
    return x.cast(slope.int32).cast(x.dtype)


@procedure_set.register()
def ceil(x: Tensor) -> Tensor:
    return (x > (b := x.trunc())).where(b + 1, b)


@procedure_set.register()
def floor(x: Tensor) -> Tensor:
    return (x < (b := x.trunc())).where(b - 1, b)


@procedure_set.register()
def square(x):
    return x * x


@procedure_set.register()
def clip(x, min_, max_):
    return x.maximum(min_).minimum(max_)


@procedure_set.register()
def abs(x):
    return x.relu() + (-x).relu()


@procedure_set.register()
def sign(x):
    return x / (x.abs() + 1e-10)


@procedure_set.register()
def reciprocal(x):
    return x.ones_like() / x


@procedure_set.register()
def matmul(x, w):
    x = x.reshape((*x.shape[0:-1], 1, x.shape[-1]))
    w = w.reshape((*w.shape[0:-2], 1, w.shape[-2], w.shape[-1])).transpose(-1, -2)
    return (x * w).sum(-1).reshape((*x.shape[0:-2], -1))


@procedure_set.register()
def getitem(x, indices) -> Tensor:
    # 1. indices normalization and validation
    # treat internal tuples and lists as Tensors and standardize indices to list type
    if isinstance(indices, list) and all(isinstance(s, int) for s in indices):
        indices = [
            slope.tensor(
                indices,
                x.device,
            )
        ]
    elif isinstance(indices, (tuple, list)):
        indices = [slope.tensor(list(i), x.device) if isinstance(i, (tuple, list)) else i for i in indices]
    else:
        indices = [indices]

    # filter ellipsis and fill with slice(None) or fill rest of indices with slice(None)
    ellipsis_idx = [dim for dim, i in enumerate(indices) if i is Ellipsis]
    fill_idx = ellipsis_idx[0] if ellipsis_idx else len(indices)
    num_indices = len(indices) - len(ellipsis_idx) - sum(1 for i in indices if i is None)
    indices[fill_idx : fill_idx + 1] = [slice(None)] * (len(x.shape) - num_indices)

    # use Dict[type, List[dimension]] to track elements in indices
    type_dim: DefaultDict[Union[type, None], List[int]] = defaultdict(list)

    # record None for dimension injection later and filter None and record rest of indices
    type_dim[None] = [dim for dim, i in enumerate(indices) if i is None]
    indices_filtered = [v for v in indices if v is not None]
    for dim, i in enumerate(indices_filtered):
        type_dim[type(i)].append(dim)

    for index_type in type_dim:
        if index_type not in [None, int, slice, Tensor]:
            raise IndexError(f"{index_type=} not supported")
    if len(ellipsis_idx) > 1:
        raise IndexError("indices can only have a single ellipsis ('...')")
    if num_indices > x.ndim:
        raise IndexError(f"too many {num_indices=} for {x.ndim=}")

    # 2. basic indexing, uses only movement ops (no copy)
    # currently indices_filtered: Tuple[Union[slice, int, Tensor], ...]
    # turn indices in indices_filtered to Tuple[shrink_arg, strides]
    for dim in type_dim[int]:
        if (index := indices_filtered[dim]) >= (size := x.shape[dim]) or index < -size:
            raise IndexError(f"{index=} is out of bounds on {dim=} with {size=}")
        indices_filtered[dim] = ((index, index + 1), 1) if index >= 0 else ((size + index, size + index + 1), 1)
    for dim in type_dim[slice]:
        if (index := indices_filtered[dim]).step == 0:
            raise ValueError(f"{index=} on {dim=} cannot have 0 as step")
        s, e, st = index.indices(x.shape[dim])
        indices_filtered[dim] = ((0, 0) if (st * (e - s)) < 0 else (s, e) if st > 0 else (e + 1, s + 1), st)
    # record tensors and skip all Tensor dims for basic indexing
    tensor_index: List[Tensor] = []
    for dim in type_dim[Tensor]:
        tensor_index.append(index := indices_filtered[dim])
        if not dtypes.is_int(index.dtype):
            raise IndexError(f"{index.dtype=} on {dim=} is not supported, only int tensor indexing is supported")
        indices_filtered[dim] = ((0, x.shape[dim]), 1)

    new_slice, strides = ((), ()) if not indices_filtered else zip(*indices_filtered)
    ret = x.padslice(new_slice).flip(tuple(i for i, s in enumerate(strides) if s < 0))
    if any(abs_(s) != 1 for s in strides):
        strides = tuple(abs_(s) for s in strides)
        round_up = lambda num, amt: (num + amt - 1) // amt * amt
        ret = ret.pad(tuple((0, round_up(sh, s) - sh) for s, sh in zip(strides, ret.shape)))
        ret = ret.reshape(tuple(flatten((sh // s, s) for s, sh in zip(strides, ret.shape))))
        ret = ret.padslice(tuple(flatten(((0, sh), (0, 1)) for sh in ret.shape[::2]))).reshape(ret.shape[::2])

    # inject 1 for dim where it's None and collapse dim for int
    new_shape = list(ret.shape)
    for dim in type_dim[None]:
        new_shape.insert(dim, 1)
    for dim in (
        tuple(dim + sum_(1 for d in type_dim[None] if dim >= d) for dim in reversed(type_dim[int]))
        # dims_collapsed := tuple(dim + sum_(1 for d in type_dim[None] if dim >= d) for dim in reversed(type_dim[int]))
    ):
        new_shape.pop(dim)

    ret = ret.reshape(new_shape)

    # 3. advanced indexing (copy)
    if type_dim[Tensor]:
        for i in tensor_index:
            while i.ndim < ret.ndim:
                i = i[None]
            ret = ret.gather_nd(i)
    return ret


@procedure_set.register()
def padslice(x, arg: Sequence[Optional[Tuple[int, int]]], value: float = 0):
    def flatten_seq(l: Iterator):
        return tuple(item for sublist in l for item in sublist)

    # some dim are pad, some are sliced
    arg_ = tuple([a if a is not None else (0, s) for s, a in zip(x.shape, arg)])
    padding = tuple([(max(0, -p[0]), max(0, p[1] - x.shape[i])) for i, p in enumerate(arg_)])
    if len(padding) == 0:
        breakpoint()
    x = x.pad(flatten_seq(padding)[::-1], value=value)  # flatten
    starts, limits, strides = tuple(zip(*[(p[0] + padding[i][0], p[1] + padding[i][0], 1) for i, p in enumerate(arg_)]))
    x = x.slice(starts, limits, strides)
    return x


@procedure_set.register()
def pad2d(x, padding: Union[List[int], Tuple[int, ...]], value: float = 0):
    # (padding_left, padding_right, padding_top, padding_bottom)
    slc = [(-p0, s + p1) for p0, p1, s in zip(padding[::2], padding[1::2], x.shape[::-1])][::-1]
    return x.padslice([(0, s) for s in x.shape[: -(len(padding) // 2)]] + slc, value=value)


@procedure_set.register()
def gather_nd(x, w, batch_dims=0):
    def gather(x, w, dim: int):
        assert w.ndim == x.ndim, "x.ndim must equal w.ndim"
        assert all(s >= i for s, i in zip(x.shape, w.shape)), "all dim of idx.shape must be smaller than x.shape"
        if dim < 0:
            dim += x.ndim
        w = w.transpose(ax=dim, aw=0).expand_dims(-1)
        permarg = list(range(x.ndim))
        permarg = permarg[1:dim] + [permarg[0]] + permarg[dim + 1 :] + [permarg[dim]] if dim != 0 else permarg[1:] + [permarg[0]]
        return (
            (
                (
                    w
                    == slope.arange(
                        x.shape[dim],
                        dtype=w.dtype,
                        device=w.device,
                    )
                ).cast(x.dtype)
                * x.permute(*permarg).padslice(tuple([*[(0, sh) for sh in w.shape[1:-1]], (0, x.shape[dim])])).expand_dims(0)
            )
            .sum(-1)
            .transpose(ax=0, aw=dim)
        )

    def _gather_nd_single(x_, w_):
        return gather(x_, w_.transpose(-1, 0), 0)

    assert batch_dims == 0
    gather_nd_ = (
        functools.reduce(lambda g, f: f(g), [slope.vmap] * int(batch_dims), _gather_nd_single) if batch_dims > 0 else _gather_nd_single
    )
    return gather_nd_(x, w)


@procedure_set.register()
def scatter_nd(params, indices, updtes):
    raise NotImplementedError


@procedure_set.register()
@staticmethod
def stack(tensors, dim=0):
    first = tensors[0].expand_dims(dim)
    expand_dimsd_tensors = [tensor.expand_dims(dim) for tensor in tensors[1:]]
    return first.cat(*expand_dimsd_tensors, dim=dim)


@procedure_set.register()
def repeat(x, repeats):
    base_shape = (1,) * (len(repeats) - x.ndim) + x.shape
    new_shape = [x for b in base_shape for x in [1, b]]
    expand_shape = [x for rs in zip(repeats, base_shape) for x in rs]
    final_shape = [r * s for r, s in zip(repeats, base_shape)]
    return x.reshape(new_shape).expand(expand_shape).reshape(final_shape)


@procedure_set.register()
def split(x, num: int, dim: int):
    dim, step = dim + x.ndim if dim < 0 else dim, math.ceil(x.shape[dim] / num)
    slice_params = [[slice(None)] * dim + [slice(k, k + step)] for k in range(0, x.shape[dim], step)]
    return tuple(x[tuple(sl)] for sl in slice_params)


@procedure_set.register()
def squeeze(x, dim=None):
    if dim is None:
        return x if 1 not in x.shape else x.reshape(*[size for size in x.shape if size != 1])
    if dim <= 0 and x.ndim == 0:
        return x  # This is to match PyTorch behavior
    if not -x.ndim <= dim < x.ndim:
        raise IndexError(
            f"Dimension out of range (expected to be in range of [{-x.ndim if x.ndim > 0 else x.ndim-1}, {x.ndim-1 if x.ndim > 0 else x.ndim}], but got {dim})"
        )
    if dim < 0:
        dim += x.ndim
    return x if x.shape[dim] != 1 else x.reshape(*[size for idx, size in enumerate(x.shape) if idx != dim])


@procedure_set.register(aliases=("expand_dims",))
def unsqueeze(x, dim) -> Tensor:
    if dim < 0:
        dim = len(x.shape) + dim + 1
    return x.reshape(x.shape[:dim] + (1,) + x.shape[dim:])


@procedure_set.register()
def transpose(x, ax=1, aw=0):
    order = list(range(len(x.shape)))
    order[ax], order[aw] = order[aw], order[ax]
    return x.permute(tuple(order))


@procedure_set.register()
def flatten(x, start_dim=0):
    return x.reshape(shape=x.shape[:start_dim] + (-1,))


@procedure_set.register()
def cumsum(x, dim: int = 0):
    return x.transpose(dim, -1).pad((x.shape[dim] - 1, 0)).pool((x.shape[dim],)).sum(-1).transpose(dim, -1)


@staticmethod
@procedure_set.register()
def arange_with_cumsum(start, stop=None, step=1):
    if stop is None:
        stop, start = start, 0
    return slope.full((math.ceil((stop - start) / step),), step).cumsum() + (start - step)


@procedure_set.register()
def one_hot(x, k, dtype=dtypes.int32):
    return (x[:, None].cast(dtype) == slope.arange(k, dtype=dtype)).cast(dtype)


@procedure_set.register()
def relu(x):
    return x.maximum(slope.zeros_like(x))


@procedure_set.register()
def leakyrelu(x, neg_slope=0.01):
    return x.relu() - (slope.full_like(x, -neg_slope) * x).relu()


@procedure_set.register()
def sigmoid(x):
    return 1 / (1 + (-x).exp())


@procedure_set.register()
def tanh(x):
    return 2.0 * ((2.0 * x).sigmoid()) - 1.0


@procedure_set.register()
def swish(x):
    return x * x.sigmoid()


@procedure_set.register()
def gelu(x):
    return 0.5 * x * (1 + (x * 0.7978845608 * (1 + 0.044715 * x * x)).tanh())


@procedure_set.register()
def linear(x, w, b=None):
    x = x @ w.transpose(-2, -1)
    return x + b[None, ...] if b is not None else x


@procedure_set.register()
def sequential(x, modules, *args, **kwargs):
    for module in modules:
        x = module(x, *args, **kwargs)
    return x


@procedure_set.register()
def pool(
    x,
    kernel_size: Tuple[int, ...],
    stride: Union[Tuple[int, ...], int] = 1,
    dilation: Union[Tuple[int, ...], int] = 1,
):
    def make_pair(x: Union[int, Tuple[int, ...]], cnt=2) -> Tuple[int, ...]:
        return (x,) * cnt if isinstance(x, int) else x

    def flatten_seq(l):
        return [item for sublist in l for item in sublist]

    k_ = kernel_size
    assert len(x.shape) >= len(k_), f"can't pool {x.shape} with {k_}"
    s_, d_ = make_pair(stride, len(k_)), make_pair(dilation, len(k_))
    assert len(k_) == len(s_) and len(k_) == len(d_), f"stride/dilation mismatch kernel:{k_} stride:{s_} dilation:{d_}"
    slc_prefix, prefix, i_ = (
        [(0, x) for x in x.shape[0 : -len(k_)]],
        x.shape[0 : -len(k_)],
        x.shape[-len(k_) :],
    )
    if any(k > s for k, s in zip(k_, s_)) or any(d != 1 for d in d_):
        o_ = [(i - d * (k - 1) - 1) // s + 1 for i, d, k, s in zip(i_, d_, k_, s_)]
        e_ = [math.ceil(k * (i + d) / i) for k, i, d in zip(k_, i_, d_)]  # expands such that we don't need padding
        xup = x
        xup = xup.reshape((*prefix, *flatten_seq((1, i) for i in i_)))
        xup = xup.expand((*prefix, *flatten_seq((e, i) for e, i in zip(e_, i_))))
        xup = xup.reshape((*prefix, *[e * i for e, i in zip(e_, i_)]))
        # slide by dilation
        xup = xup.padslice(slc_prefix + [(0, k * (i + d)) for k, i, d in zip(k_, i_, d_)])
        xup = xup.reshape((*prefix, *flatten_seq((k, i + d) for k, i, d in zip(k_, i_, d_))))
        xup = xup.padslice(slc_prefix + flatten_seq(((0, k), (0, o * s)) for k, o, s in zip(k_, o_, s_)))
        # handle stride, and permute to move reduce to the end
        xup = xup.reshape((*prefix, *flatten_seq((k, o, s) for k, o, s in zip(k_, o_, s_))))
        xup = xup.padslice(slc_prefix + flatten_seq(((0, k), (0, o), (0, 1)) for k, o in zip(k_, o_)))
        xup = xup.reshape((*prefix, *flatten_seq((k, o) for k, o in zip(k_, o_))))
        return xup.permute(
            (
                *range(len(prefix)),
                *[len(prefix) + i * 2 + 1 for i in range(len(k_))],
                *[len(prefix) + i * 2 for i in range(len(k_))],
            )
        )
    o_ = [(i + (s - k)) // s for i, s, k in zip(i_, s_, k_)]
    xup = x.padslice(slc_prefix + [(0, o * s) for o, s in zip(o_, s_)])
    xup = xup.reshape((*prefix, *flatten_seq(((o, s) for o, s in zip(o_, s_)))))
    xup = xup.padslice((slc_prefix + flatten_seq(((0, o), (0, k)) for o, k in zip(o_, k_))))
    return xup.permute(
        (
            *range(len(prefix)),
            *[len(prefix) + i * 2 for i in range(len(k_))],
            *[len(prefix) + i * 2 + 1 for i in range(len(k_))],
        )
    )


@procedure_set.register()
def avgpool2d(x, kernel_size=(2, 2), stride=None):
    def make_pair(x: Union[int, Tuple[int, ...]], cnt=2) -> Tuple[int, ...]:
        return (x,) * cnt if isinstance(x, int) else x

    return x.pool(make_pair(kernel_size), stride if stride is not None else kernel_size).mean(
        dim=tuple(range(0 - len(make_pair(kernel_size)), 0))
    )


@procedure_set.register()
def maxpool2d(x, kernel_size=(2, 2), stride=None, dilation=1):
    def make_pair(x: Union[int, Tuple[int, ...]], cnt=2) -> Tuple[int, ...]:
        return (x,) * cnt if isinstance(x, int) else x

    return x.pool(
        make_pair(kernel_size),
        stride if stride is not None else kernel_size,
        dilation,
    ).max(dim=tuple(range(0 - len(make_pair(kernel_size)), 0)))


@procedure_set.register()
def conv(x, w, groups=1, stride=1, dilation=1, padding=0):
    (bs, cin_), (cout, cin), D = x.shape[:2], w.shape[:2], w.shape[2:]
    assert groups * cin == cin_ and len(x.shape) == len(
        w.shape
    ), f"Input dim shape {x.shape} does not match the shape of the ws {w.shape}. ({groups*cin} vs. {cin_})"
    if isinstance(padding, (tuple, list)):
        assert len(padding) == 2 * len(D) or len(padding) == len(
            D
        ), f"Expected padding of length {2*len(D)} or {len(D)}, but got {len(padding)} for tensor of shape {x.shape}"
    padding_ = (
        [padding] * 2 * len(D)
        if isinstance(padding, int)
        else (padding if len(padding) == 2 * len(D) else [p for p in padding for _ in range(2)][::-1])
    )
    padding_ = tuple(padding_)

    def pad2d(x, padding: Union[List[int], Tuple[int, ...]], value: float = 0):
        # (padding_left, padding_right, padding_top, padding_bottom)
        slc = [(-p0, s + p1) for p0, p1, s in zip(padding[::2], padding[1::2], x.shape[::-1])][::-1]
        return x.padslice([(0, s) for s in x.shape[: -(len(padding) // 2)]] + slc, value=value)

    x = pad2d(x, padding_)
    x = x.pool(D, stride, dilation)  # (bs, groups*cin, oy, ox, H, W)
    rcout, oyx = cout // groups, x.shape[2 : -len(D)]
    x = x.reshape((bs, groups, cin, 1, *oyx, *D))
    x = x.expand((bs, groups, cin, rcout, *oyx, *D))
    x = x.permute(
        (
            0,
            1,
            3,
            *[4 + i for i in range(len(oyx))],
            2,
            *[4 + len(oyx) + i for i in range(len(D))],
        )
    )
    # (bs, groups, rcout, *oyx, cin, *D)
    x = x * w.reshape((1, groups, rcout, *[1] * len(oyx), cin, *D))
    x = x.sum([-1 - i for i in range(1 + len(oyx))], keepdim=True)
    x = x.reshape((bs, cout, *oyx))
    ret = x
    return ret


@procedure_set.register()
def conv_transpose(x, w, groups=1, stride=1, dilation=1, padding=0, output_padding=0):
    make_pair = lambda x, cnt=2: (x,) * cnt if isinstance(x, int) else x
    flatten_seq = lambda l: [item for sublist in l for item in sublist]
    D, trailing = w.shape[2:], tuple(range(3, len(w.shape) + 1))
    w = w.reshape(((groups, w.shape[0] // groups, w.shape[1], *w.shape[2:])))  # (1, 64, 64, 3, 3)
    w = w.permute((0, 2, 1, *trailing))
    w = w.flip(trailing)
    stride = make_pair(stride, len(D))
    if any(s > 1 for s in stride):
        x = x.reshape((*x.shape[:2], *flatten_seq((k, 1) for k in x.shape[2:])))
        pads = (0, 0, 0, 0, *flatten_seq((0, 0, 0, s - 1) for s in stride))
        pads = pads[::-1]
        x = x.pad(pads)
        x = x.reshape(*x.shape[:2], *[k * s for k, s in zip(x.shape[2::2], stride)])
        x = x.slice(
            (0, 0) + (0,) * len(x.shape[2:]),
            (x.shape[0], x.shape[1]) + tuple([k - (s - 1) for k, s in zip(x.shape[2:], stride)])[::-1],
        )
    padding = flatten_seq(
        (
            ((k - 1) * d - p, (k - 1) * d - p + op)
            for k, d, p, op in reversed(
                list(
                    zip(
                        D,
                        make_pair(dilation, len(D)),
                        make_pair(padding, len(D)),
                        make_pair(output_padding, len(D)),
                    )
                )
            )
        )
    )
    w = w.reshape((w.shape[0] * w.shape[1], *w.shape[2:]))
    return x.conv(w, groups=groups, dilation=dilation, padding=padding)


@procedure_set.register()
def batchnorm(x, weight, bias, mean, invstd):
    broadcast_shape = (1, -1) + (1,) * len(x.shape[2:])
    x = (x - mean.reshape(broadcast_shape)) * invstd.reshape(broadcast_shape)
    if weight is not None and bias is not None:
        x = x * weight.reshape(broadcast_shape) + bias.reshape(broadcast_shape)
    return x


@procedure_set.register()
def layernorm(x, dim=-1, eps: float = 1e-5) -> Tensor:
    y = x - x.mean(dim, keepdim=True)
    return y.mul((y * y).mean(dim, keepdim=True).add(eps).rsqrt())


@procedure_set.register()
def dropout(x, p, training=False) -> Tensor:
    if not training or p == 0:
        return x
    mask = (slope.rand(*x.shape, device=x.device) >= p).cast(slope.bool)
    return x * mask * (1 / (1.0 - p))


@procedure_set.register()
def scaled_dot_product_attention(
    x,
    key: Tensor,
    value: Tensor,
    attn_mask: Optional[Tensor] = None,
    dropout_p: float = 0.0,
    is_causal: bool = False,
) -> Tensor:
    if is_causal:
        attn_mask = slope.ones(x.shape[-2], key.shape[-2], device=x.device).tril(0).cast(slope.bool)
    if attn_mask is not None and attn_mask.dtype == slope.bool:
        attn_mask = (attn_mask == 0).where(-float("inf"), attn_mask)
    return (x @ key.transpose(-2, -1) / math.sqrt(x.shape[-1]) + attn_mask).softmax(-1).dropout(dropout_p) @ value


@procedure_set.register()
def binary_cross_entropy(x, y: Tensor) -> Tensor:
    return (-y * x.log() - (1 - y) * (1 - x).log()).mean()


@procedure_set.register()
def binary_cross_entropy_with_logits(x, y: Tensor) -> Tensor:
    return (x.maximum(0) - y * x + (1 + x.abs().__neg__().exp()).log()).mean()


# @procedure_set.register()
# def cross_entropy(x, y, ignore_index=-1) -> Tensor:
#     loss_mask = (y != ignore_index).reshape(-1, 1)
#     y_counter = slope.arange(x.shape[-1], dtype=slope.int32)[None, ..., None]
#     y_oh = (y_counter == y[..., None, None]).where(-1.0, 0.0).squeeze(-1)
#     y = y * loss_mask
#     return (x.log_softmax(-1) * y_oh).sum()  / loss_mask.sum()


@procedure_set.register()
def cross_entropy(x, y) -> Tensor:
    y_counter = slope.arange(x.shape[-1], dtype=slope.int32)[None, ..., None]
    y_oh = (y_counter == y[..., None, None]).where(-1.0, 0.0).squeeze(-1)
    return (x.log_softmax(-1) * y_oh).sum()


@procedure_set.register()
def softmax(x, dim=-1):
    m = x - x.max(dim, keepdim=True)
    e = m.exp()
    ss = e.sum(dim, keepdim=True)
    return e / ss


@procedure_set.register()
def log_softmax(x, dim=-1):
    x = x - x.max(dim, keepdim=True)
    logsumexp_x = x.exp().sum(dim, keepdim=True).log()
    return x - logsumexp_x


# TODO:

# nograd_functions = [
#     np.all,
#     np.allclose,
#     np.any,
#     #np.argmax,
#     np.argmin,
#     np.argpartition,
#     np.argsort,
#     np.argwhere,
#     np.around,
#     np.array_equal,
#     np.array_equiv,
#     np.ceil,
#     np.count_nonzero,
#     #np.equal,
#     np.fix,
#     np.flatnonzero,
#     np.floor,
#     np.floor_divide,
#     np.greater,
#     np.greater_equal,
#     np.isclose,
#     np.isfinite,
#     np.isinf,
#     np.isnan,
#     np.isneginf,
#     np.isposinf,
#     np.isscalar,
#     np.less,
#     np.less_equal,
#     np.logical_and,
#     np.logical_not,
#     np.logical_or,
#     np.logical_xor,
#     np.ndim,
#     np.nonzero,
#     np.not_equal,
#     np.ones_like,
#     np.result_type,
#     np.rint,
#     np.round,
#     np.searchsorted,
#     np.shape,
#     np.sign,
#     np.size,
#     np.trunc,
#     np.zeros_like,
# ]



import slope
from slope import nn
import math


class Net(nn.Module):
    def __init__(self, in_dim, hid_dim, out_dim, act=nn.ReLU()):
        self.flatten_fn = nn.Fn(lambda x: x.reshape(shape=(x.shape[0], math.prod(x.shape[1:]))))
        self.linear1 = nn.Linear(in_dim, hid_dim)
        self.linear2 = nn.Linear(hid_dim, out_dim)
        self.act = act

    def __call__(self, x):
        x = self.flatten_fn(x)
        x = self.linear1(x)
        x = self.act(x)
        x = self.linear2(x)
        return x


model = Net(784, 100, 10)


@slope.jit
def train_step(model, batch, optimizer):
    def train_loss_fn(model, batch):
        x, y = batch
        logits = model(x)
        loss = logits.cross_entropy(y) / x.size(0)

        return loss, (model, logits)

    (loss, (model, logits)), gmodel = slope.value_and_grad(train_loss_fn, has_aux=True)(model, batch)
    model, optimizer = optimizer(model, gmodel)
    return loss, logits, model, optimizer, gmodel


optimizer = slope.nn.SGD(model, lr=0.1, momentum=0.9, weight_decay=1e-4)
N = 100
y = slope.arange(10, dtype=slope.float32)[None].expand((N // 10, 10)).reshape(-1)
x = slope.randn(N, 784)
train_step(model, (x, y), optimizer)
with slope.core.Profiling(f"RUN"):
    loss, logits, model, optimizer, gmodel = train_step(model, (x, y), optimizer)



import slope
import slope.core
from slope.core import (
    Backend,
    Operator,
    Instruction,
    Program,
    MetaOperator,
    ProcedureSet,
    Tensor,
    TensorBuffer,
    SymbolicTensor,
    list_map,
    dtypes,
    devices,
    DType,
    Device,
    CodegenOutput,
)

import math
import numpy as np
from typing import (
    List,
    Any,
)
import os

from slope.operators import operator_set
from slope.procedures import procedure_set
import tempfile
import importlib
import pygments
from pygments.lexers.python import PythonLexer
from pygments.formatters import Terminal256Formatter


def annotate_shape(symval):
    xdtype = symval.dtype.mlir
    if len(symval.shape) > 0:
        xshape = f"{'x'.join((repr(i) for i in symval.shape))}"
        return f"tensor<{xshape}x{xdtype}>"
    else:
        return f"tensor<{xdtype}>"


def annotate_sig(in_symvals, out_symvals):
    if isinstance(in_symvals, SymbolicTensor):
        in_symvals = (in_symvals,)
    if isinstance(out_symvals, SymbolicTensor):
        out_symvals = (out_symvals,)
    in_sig = f"({','.join(annotate_shape(t) for t in in_symvals)})"
    out_sig = f"({','.join(annotate_shape(t) for t in out_symvals)})"
    sig = f"{in_sig} -> {out_sig}"
    return sig


class NumpyBackend(Backend):
    dtype_for_indices = dtypes.int64
    dtype_map = {
        dtypes.float32: np.dtypes.Float32DType(),
        dtypes.uint8: np.dtypes.UInt8DType(),
        dtypes.int8: np.dtypes.Int8DType(),
        dtypes.bool: np.dtypes.BoolDType(),
        dtypes.int32: np.dtypes.Int32DType(),
        dtypes.int64: np.dtypes.Int64DType(),
        dtypes.uint64: np.dtypes.UInt64DType(),
        dtypes.float16: np.dtypes.Float16DType(),
    }
    device_map = {
        devices.cpu: "cpu",
    }

    dtype_map_inv = {v: k for k, v in dtype_map.items()}
    device_map_inv = {v: k for k, v in device_map.items()}

    def from_numpy(self, val, dtype=None, device=None):
        dtype = dtype or self.DEFAULT_DTYPE
        return Tensor(TensorBuffer(np.array(val, dtype=self.dtype_map[dtype])))

    def numpy_of(self, tensor: Tensor, memmap=False):
        if not memmap:
            return tensor.buf.val
        with tempfile.NamedTemporaryFile() as arr_file:
            arr = np.memmap(arr_file.name, dtype=tensor.dtype.numpy, shape=tensor.shape, mode="w+")
            arr[:] = tensor.buf.val
            return arr

    def shape_of(self, tensor):
        return tuple(int(i) for i in tensor.buf.val.shape)

    def dtype_of(self, tensor: Tensor):
        return self.dtype_map_inv[tensor.buf.val.dtype]

    def device_of(self, tensor: Tensor):
        return devices.cpu

    def codegen(self, program: Program, args, *, fn_name: str = "main", fn_defs=dict()) -> List[Any]:
        if fn_name == "main":
            assert not hasattr(self, "fn_count")
            self.fn_count = 0

        def indent(code):
            spaces = " " * (len(code) - len(code.lstrip()))
            spaces += " " * 4
            return "\n".join([spaces + line for line in code.strip().split("\n")])

        # codegen is recursive if jit-of-jit happens
        body_code_lines = []

        for instruction in program.instructions:
            if len(instruction.out_binders) == 0:  # skip codegen for function returns nothing
                continue
            in_vals = list_map(lambda x: program.env[x], instruction.inputs)
            out_vals = list_map(lambda z: program.env[z], instruction.out_binders)
            if isinstance(instruction.op, MetaOperator):
                impl_code, fn_defs = self.impls[instruction.op](args, instruction, fn_defs, in_vals, out_vals)
            else:
                if instruction.op in self.impls.keys():
                    impl_code = self.impls[instruction.op](*in_vals, *out_vals, **instruction.params)
                else:
                    # No impl is defined, fallback to procedure
                    impl_code, fn_defs = self.codegen_impl_as_procedure(args, instruction, fn_defs, in_vals, out_vals)

            for impl_code_line in impl_code.split("\n"):  # handle multi-line code
                body_code_lines += [indent(impl_code_line)]

        in_binders = list_map(lambda x: program.env[x], program.in_binders)
        arg_type_strs = [f"{inb.name}" for inb in in_binders]
        fn_args_str = ", ".join(arg_type_strs)

        functions_code_lines = []
        if fn_name == "main":
            for fn_def_code_lines in fn_defs.values():
                functions_code_lines += [indent(line) for line in fn_def_code_lines]

        outs = list_map(lambda x: program.env[x], program.outs)
        out_type_strs = [f"{out.name}" for out in outs]
        out_type_str = ", ".join(out_type_strs)
        head_code_line = [f"def {fn_name}({fn_args_str}): # ({out_type_str})"]
        return_line = [indent(f"return {out_type_str}")]
        code_lines = head_code_line + functions_code_lines + body_code_lines + return_line
        if slope.LOG_JIT:
            formatted_code = pygments.highlight("\n".join(code_lines), PythonLexer(), Terminal256Formatter())
            slope.core.dblog(f"\n---- {program.name} codegen:\n\n" + formatted_code + "\n\n===============\n")

        if fn_name == "main":
            del self.fn_count
        assert len(outs) == len(program.outs)
        return CodegenOutput(
            code_lines=code_lines,
            fn_defs=fn_defs,
            in_binders=in_binders,
            outs=outs,
        )

    def codegen_impl_as_procedure(self, args, instruction: Instruction, fn_defs, in_vals, out_vals):
        op: Operator = instruction.op
        op_name = {v: k for k, v in vars(self.operator_set).items()}[op]
        op_procedure = getattr(self.procedure_set, op_name)
        symvals_in = tuple(inp.symval for inp in instruction.inputs)
        params = instruction.params
        op_program, consts, _ = slope.core.make_program(
            op_procedure,
            *symvals_in,
            static_args=tuple(params.items()),
            name=op.name,
        )
        name = slope.core.jit.get_jit_name(tuple(symvals_in), params)
        if name not in fn_defs.keys():
            op_codegen_output: CodegenOutput = self.codegen(
                op_program,
                args,
                fn_name=name,
                fn_defs=fn_defs,
            )
            fn_defs = {**fn_defs, **op_codegen_output.fn_defs}
            fn_defs[name] = op_codegen_output.code_lines
        in_names = ", ".join(i.name for i in in_vals)
        out_names = ", ".join(o.name for o in out_vals)
        impl_code = f"{out_names} = {name}({in_names})"
        return impl_code, fn_defs

    def compile(self, codegen_output):
        deps_dict = dict()
        deps_dict["numpy"] = importlib.import_module("numpy")
        deps_dict["np"] = deps_dict["numpy"]
        deps_dict["math"] = importlib.import_module("math")
        code_lines = codegen_output.code_lines
        exec_locals = dict()
        code = "\n".join(code_lines)
        exec(compile(code, "<string>", "exec"), deps_dict, exec_locals)
        fn = exec_locals["main"]
        return fn, code

    def export(self, jit_output: slope.core.JitOutput, output_path, export_params, input_names, output_names, **kwargs):
        os.makedirs(output_path, exist_ok=True)
        consts_dir_path = os.path.join(output_path, "consts")
        os.makedirs(consts_dir_path, exist_ok=True)
        in_binders = jit_output.codegen_output.in_binders
        outs = jit_output.codegen_output.outs
        num_consts = jit_output.program.num_consts
        load_consts_code = ""
        for i in range(num_consts):
            const_name = in_binders[i].name
            const_path = os.path.join(consts_dir_path, f"{const_name}.npy")
            load_consts_code += f"""{const_name} = np.load(os.path.join(consts_dir_path, "{const_name}.npy"))\n"""
            np.save(const_path, in_binders[i].symval.dtype.numpy)
        input_args_code = ", ".join(inb.name for inb in in_binders[num_consts:])
        args_code = ", ".join(inb.name for inb in in_binders)
        input_arg_names = [inb.name for inb in in_binders[num_consts:]]
        input_arg_names_str = ", ".join(input_arg_names)
        outs_names = [out.name for out in outs]

        test_input_code = ""
        for i in range(num_consts, len(in_binders)):
            input_name = in_binders[i].name
            input_shape = in_binders[i]["type"].shape
            dtype = in_binders[i]["type"].dtype
            input_dtype = f"{'np.' if dtype is not dtypes.bool else ''}{self.dtype_map(dtype)}"
            test_input_code += f"""    {input_name} = np.ones({input_shape}, dtype={input_dtype})\n"""

        module_path = os.path.join(output_path, "__init__.py")
        module_code = f"""import numpy as np
    import os
    root_path = os.path.dirname(__file__)
    consts_dir_path =  os.path.join(root_path, "consts")
    {load_consts_code}
    {jit_output.code}

    input_arg_names = {input_arg_names}
    out_names = {outs_names}

    def run({input_args_code}):
        return main({args_code})

    if __name__ == "__main__":
    {test_input_code}
        for inp_name, inp in zip(input_arg_names, ({input_arg_names_str})):
            print(f"{{inp_name}} = ")
            print(inp)
            print(f"dtype: {{inp.dtype}}")
            print(f"shape: {{inp.shape}}")
            print()

        outs = run({input_arg_names_str})

        print("outputs:")
        for out_name, out in zip(out_names, outs):
            print(f"{{out_name}} = ")
            print(out)
            print(f"dtype: {{out.dtype}}")
            print(f"shape: {{out.shape}}")
            print()
    """
        with open(module_path, "w") as f:
            f.write(module_code)
            slope.dblog(module_code, enable=slope.LOG_JIT)


backend = NumpyBackend(operator_set, procedure_set)


@backend.set_impl(backend.operator_set.jit_op)
def jit_op_impl(self, args, instruction, fn_defs, in_vals, out_vals):
    jit_program = instruction.params["program"]
    jit_name = f"{jit_program.name}"
    if jit_name not in fn_defs.keys():
        jit_codegen_output = self.codegen(
            jit_program,
            args,
            fn_name=jit_name,
            fn_defs=fn_defs,
        )
        fn_defs[jit_name] = jit_codegen_output.code_lines
        fn_defs = {**fn_defs, **jit_codegen_output.fn_defs}
    args_str = ", ".join(i.name for i in in_vals)
    impl_code = f"{', '.join(o.name for o in out_vals)} = slope.{jit_name}({args_str})"
    return impl_code, fn_defs


### Operator Impls


@backend.set_impl(operator_set.cast)
def cast_impl(self, x, y, *, dtype):
    return f"{y.name} = {x.name}.astype({'np.' if dtype is not dtypes.bool else ''}{self.dtype_map[dtype]})"


@backend.set_impl(backend.operator_set.stop_gradient)
def stop_gradient_impl(self, x, y):
    return f"{y.name} = {x.name}"


@backend.set_impl(backend.operator_set.sqrt)
def sqrt_impl(self, x, y):
    return f"{y.name} = np.sqrt({x.name})"


@backend.set_impl(backend.operator_set.exp)
def exp_impl(self, x, y):
    return f"{y.name} = np.exp({x.name})"


@backend.set_impl(backend.operator_set.log)
def log_impl(self, x, y):
    return f"{y.name} = np.log({x.name})"


@backend.set_impl(backend.operator_set.sin)
def sin_impl(self, x, y):
    return f"{y.name} = np.sin({x.name})"


@backend.set_impl(backend.operator_set.invert)
def invert_impl(self, x, y):
    return f"{y.name} = ~{x.name}"


@backend.set_impl(backend.operator_set.add)
def add_impl(self, x, w, y):
    return f"{y.name} = {x.name} + {w.name}"


@backend.set_impl(backend.operator_set.sub)
def sub_impl(self, x, w, y):
    return f"{y.name} = {x.name} - {w.name}"


@backend.set_impl(backend.operator_set.mul)
def mul_impl(self, x, w, y):
    return f"{y.name} = {x.name} * {w.name}"


@backend.set_impl(backend.operator_set.div)
def div_impl(self, x, w, y):
    return f"{y.name} = {x.name} / {w.name}"


@backend.set_impl(backend.operator_set.pow)
def pow_impl(self, x, w, y):
    return f"{y.name} = {x.name} ** {w.name}"


@backend.set_impl(backend.operator_set.equal)
def equal_impl(self, x, w, y):
    return f"{y.name} = {x.name} == {w.name}"


@backend.set_impl(backend.operator_set.less)
def less_impl(self, x, w, y):
    return f"{y.name} = {x.name} < {w.name}"


@backend.set_impl(backend.operator_set.greater)
def greater_impl(self, x, w, y):
    return f"{y.name} = {x.name} > {w.name}"


@backend.set_impl(backend.operator_set.maximum)
def maximum_impl(self, x, w, y):
    return f"{y.name} = np.maximum({x.name}, {w.name})"


@backend.set_impl(backend.operator_set.matmul)
def matmul_impl(self, x, w, y):
    return f"{y.name} = {x.name} @ {w.name}"


# @backend.set_impl(backend.operator_set.where)
# def where_impl(self, x, w, u, y):
#     return f"""{y.name} = np.where({x.name}, %{w.name}, %{u.name})"""


@backend.set_impl(operator_set.sum)
def sum_impl(self, x, y, *, dim, keepdim):
    return f"{y.name} = np.sum({x.name}, axis={dim}, keepdims={keepdim})"


@backend.set_impl(operator_set.max)
def max_impl(self, x, y, *, dim, keepdim):
    return f"{y.name} = np.max({x.name}, axis={dim}, keepdims={keepdim})"


@backend.set_impl(operator_set.arange)
def arange_impl(self, y, *, start, stop, stride, dtype, device):
    return f"{y.name} = np.arange({start}, {stop}, {stride}, dtype={'np.' if dtype is not dtypes.bool else ''}{self.dtype_map[dtype]})"


@backend.set_impl(operator_set.full)
def full_impl(self, y, *, shape, fill_value, dtype, device):
    return f"{y.name} = np.full({shape}, {fill_value}, dtype={'np.' if dtype is not dtypes.bool else ''}{self.dtype_map[dtype]})"


@backend.set_impl(operator_set.random_uniform)
def random_uniform_impl(self, y, *, shape, dtype, device):
    if len(shape) > 0:
        return f"{y.name} = {'np.array(' if shape == () else ''}np.random.uniform(low=np.zeros(shape={shape})){')' if shape == () else ''}.astype(dtype={'np.' if dtype is not dtypes.bool else ''}{self.dtype_map[dtype]})"


@backend.set_impl(operator_set.random_normal)
def random_normal_impl(self, y, *, shape, dtype, device):
    return f"{y.name} = {'np.array(' if shape == () else ''}np.random.normal(loc=np.zeros(shape={shape})){')' if shape == () else ''}.astype(dtype={'np.' if dtype is not dtypes.bool else ''}{self.dtype_map[dtype]})"


@backend.set_impl(operator_set.pad)
def pad_impl(self, x, *, padding, mode, value):
    pad_width = [(lo, hi) for lo, hi in zip(padding[0::2], padding[1::2])]
    return f"ret = np.pad({x}, {pad_width}, constant_values={value})"


@backend.set_impl(operator_set.expand)
def expand_impl(self, x, y, *, shape):
    return f"{y.name} = np.broadcast_to({x.name}, shape={shape})"


@backend.set_impl(operator_set.reshape)
def reshape_impl(self, x, y, *, shape):
    return f"{y.name} = np.reshape({x.name}, newshape={shape})"


@backend.set_impl(operator_set.pad)
def pad_impl(self, x, y, *, padding, mode, value):
    padding = padding[::-1]
    pad_width = [(lo, hi) for lo, hi in zip(padding[0::2], padding[1::2])]
    return f"{y.name} = np.pad({x.name}, {pad_width}, constant_values={value})"


@backend.set_impl(operator_set.slice)
def slice_impl(self, x, y, *, starts, limits, strides):
    return f"{y.name} = {x.name}[tuple(slice(s, l, st) for s, l, st in zip({starts}, {limits}, {strides}))]"
    # slice_strs = [f"{s}:{e}:{st}" for s, e ,st in zip(starts, limits, strides)]
    # slices = ", ".join([f"np.s_[{sl}]" for sl in slice_strs])
    # return f"{y.name} = {x.name}[{slices}]"


@backend.set_impl(operator_set.cat)
def cat_impl(self, *xs, dim):
    xs, y = xs[:-1], xs[-1]
    return f"{y.name} = np.concatenate(({', '.join([x.name for x in xs])}), axis={dim})"


@backend.set_impl(operator_set.permute)
def permute_impl(self, x, y, *, perm):
    return f"{y.name} = np.transpose({x.name}, axes={perm})"


@backend.set_impl(operator_set.flip)
def flip_impl(self, x, y, *, dim):
    return f"{y.name} = np.flip({x.name}, axis={dim})"



class Gather(ShapeOp):
    get_impl = lambda: slope.RT.compiler.GatherImpl

    @staticmethod
    def run(x, idx, *, dim):
        return [x.gather(idx)]

    @staticmethod
    def _gather_batching_rule(
        dim_size,
        vals_in,
        dims_in,
        # batched_args,
        # batch_dims,
        *,
        dimension_numbers,
        slice_sizes,
        unique_indices,
        indices_are_sorted,
        mode,
        fill_value,
    ):
        operand, indices = batched_args
        operand_bdim, indices_bdim = batch_dims

        if operand_bdim is not None and indices_bdim is None:
            operand = batching.movedim(operand, operand_bdim, 0)
            slice_sizes = (operand.shape[0],) + slice_sizes
            offset_dims = (0,) + tuple(np.add(1, dimension_numbers.offset_dims))
            collapsed_slice_dims = tuple(np.add(1, dimension_numbers.collapsed_slice_dims))
            start_index_map = tuple(np.add(1, dimension_numbers.start_index_map))
            dnums = GatherDimensionNumbers(
                offset_dims=offset_dims,
                collapsed_slice_dims=collapsed_slice_dims,
                start_index_map=start_index_map,
            )
            return (
                gather(
                    operand,
                    indices,
                    dimension_numbers=dnums,
                    slice_sizes=slice_sizes,
                    unique_indices=unique_indices,
                    indices_are_sorted=indices_are_sorted,
                    mode=mode,
                    fill_value=fill_value,
                ),
                0,
            )

        elif operand_bdim is None and indices_bdim is not None:
            indices = batching.movedim(indices, indices_bdim, 0)
            offset_dims = tuple(1 + d for d in dimension_numbers.offset_dims)
            dnums = GatherDimensionNumbers(
                offset_dims=offset_dims,
                collapsed_slice_dims=dimension_numbers.collapsed_slice_dims,
                start_index_map=dimension_numbers.start_index_map,
            )
            # If batching indexed accesses into the same tensor, the batched gather may
            # no longer have sorted or unique indices.
            return (
                gather(
                    operand,
                    indices,
                    dimension_numbers=dnums,
                    slice_sizes=slice_sizes,
                    unique_indices=False,
                    indices_are_sorted=False,
                    mode=mode,
                    fill_value=fill_value,
                ),
                0,
            )

        else:
            # move batch dimensions to the front to simplify logic
            operand = batching.movedim(operand, operand_bdim, 0)
            indices = batching.movedim(indices, indices_bdim, 0)

            # This slightly awkward special case is needed because the shape rule for
            # gather does not allow size-1 slices out of a size-0 dimension, even if
            # the number of slices is zero. Likely the best fix would be to change the
            # definition of gather() so it can be batched without the construction of
            # an explicit iota of size-1 slices.
            if core.symbolic_equal_dim(operand.shape[0], 0):
                output_shape = _gather_shape_rule(
                    core.ShapedTensor(operand.shape[1:], operand.dtype),
                    core.ShapedTensor(indices.shape[1:], dtypes.canonicalize_dtype(indices.dtype)),
                    dimension_numbers=dimension_numbers,
                    slice_sizes=slice_sizes,
                    unique_indices=unique_indices,
                    indices_are_sorted=indices_are_sorted,
                    mode=mode,
                    fill_value=fill_value,
                )
                return lax.full((0,) + output_shape, lax._zero(operand)), 0

            # Example: user code had indices shape (3, 4, 5), and we have to deal with
            # indices shape (7, 3, 4, 5). We transform that to indices of shape
            # (7, 3, 4, 6) where we catd an iota that counts along our batch
            # dimension to the front of the ndindex.
            count_shape = list(indices.shape)
            count_shape[-1] = 1
            counts = lax.broadcasted_iota(indices.dtype, tuple(count_shape), 0)
            indices = lax.cat([counts, indices], len(count_shape) - 1)

            slice_sizes = (1,) + slice_sizes
            collapsed_slice_dims = (0,) + tuple(np.add(1, dimension_numbers.collapsed_slice_dims))
            offset_dims = tuple(np.add(1, dimension_numbers.offset_dims))
            start_index_map = (0,) + tuple(np.add(1, dimension_numbers.start_index_map))

            dnums = GatherDimensionNumbers(
                offset_dims=offset_dims,
                collapsed_slice_dims=collapsed_slice_dims,
                start_index_map=start_index_map,
            )
            return (
                gather(
                    operand,
                    indices,
                    dimension_numbers=dnums,
                    slice_sizes=slice_sizes,
                    unique_indices=unique_indices,
                    indices_are_sorted=indices_are_sorted,
                    mode=mode,
                    fill_value=fill_value,
                ),
                0,
            )

    @staticmethod
    def jvp(
        primals,
        tangents,
        indices,
        # g,
        # operand,
        *,
        dimension_numbers,
        slice_sizes,
        unique_indices,
        indices_are_sorted,
        mode,
        fill_value,
    ):
        return gather(
            g,
            indices,
            dimension_numbers,
            slice_sizes,
            unique_indices=unique_indices,
            indices_are_sorted=indices_are_sorted,
            mode=mode,
            fill_value=0,
        )

    @staticmethod
    def typecheck(
        x: SymbolicTensor,
        idx,
        *,
        dimension_numbers,
        slice_sizes,
        unique_indices,
        indices_are_sorted,
        mode,
        fill_value,
    ) -> List[SymbolicTensor]:
        offset_dims = dimension_numbers.offset_dims
        collapsed_slice_dims = dimension_numbers.collapsed_slice_dims
        start_index_map = dimension_numbers.start_index_map

        # Note: in JAX, index_vector_dim is always computed as below, cf. the
        # documentation of the GatherDimensionNumbers class.
        index_vector_dim = _rank(indices) - 1

        # This case should never happen in JAX, due to the implicit construction of
        # index_vector_dim, but is included for completeness.
        if _rank(indices) < index_vector_dim or index_vector_dim < 0:
            raise TypeError(
                f"Gather index leaf dimension must be within [0, rank("
                f"indices) + 1). rank(indices) is {_rank(indices)} and "
                f"gather index leaf dimension is {index_vector_dim}."
            )

        expanded_indices_shape = list(indices.shape)

        # This case should never happen in JAX, due to the implicit construction of
        # index_vector_dim, but is included for completeness.
        if len(expanded_indices_shape) == index_vector_dim:
            expanded_indices_shape.append(1)

        # Start ValidateGatherDimensions
        # In the error messages output by XLA, "offset_dims" is called "Output window
        # dimensions" in error messages. For consistency's sake, our error messages
        # stick to "offset_dims".
        _is_sorted(offset_dims, "gather", "offset_dims")
        _no_duplicate_dims(offset_dims, "gather", "offset_dims")

        output_offset_dim_count = len(offset_dims)
        output_shape_rank = len(offset_dims) + _rank(indices) - 1

        for i in range(output_offset_dim_count):
            offset_dim = offset_dims[i]
            if offset_dim < 0 or offset_dim >= output_shape_rank:
                raise TypeError(
                    f"Offset dimension {i} in gather op is out of bounds; "
                    f"got {offset_dim}, but should have been in "
                    f"[0, {output_shape_rank})"
                )

            if len(start_index_map) != indices.shape[index_vector_dim]:
                raise TypeError(
                    f"Gather op has {len(start_index_map)} elements in "
                    f"start_index_map and the bound of dimension "
                    f"{index_vector_dim=} of indices is "
                    f"{indices.shape[index_vector_dim]}. These two "
                    f"numbers must be equal."
                )

            for i in range(len(start_index_map)):
                operand_dim_for_start_index_i = start_index_map[i]
                if operand_dim_for_start_index_i < 0 or operand_dim_for_start_index_i >= _rank(operand):
                    raise TypeError(
                        f"Invalid start_index_map; domain is "
                        f"[0, {_rank(operand)}), got: "
                        f"{i}->{operand_dim_for_start_index_i}."
                    )

            _no_duplicate_dims(start_index_map, "gather", "start_index_map")

        # _is_sorted and _sorted_dims_in_range are checked in the opposite order
        # compared to the XLA implementation. In cases when the input is not sorted
        # AND there are problematic collapsed_slice_dims, the error message will thus
        # be different.
        _is_sorted(collapsed_slice_dims, "gather", "collapsed_slice_dims")
        _sorted_dims_in_range(collapsed_slice_dims, _rank(operand), "gather", "collapsed_slice_dims")
        _no_duplicate_dims(collapsed_slice_dims, "gather", "collapsed_slice_dims")
        # End ValidateGatherDimensions

        if _rank(operand) != len(slice_sizes):
            raise TypeExrror(
                f"Gather op must have one slice size for every input "
                f"dimension; got: len(slice_sizes)={len(slice_sizes)}, "
                f"input_shape.rank={_rank(operand)}"
            )

        if len(slice_sizes) != len(offset_dims) + len(collapsed_slice_dims):
            raise TypeError(
                f"All components of the offset index in a gather op must "
                f"either be a offset dimension or explicitly collapsed; "
                f"got len(slice_sizes)={len(slice_sizes)}, "
                f"output_slice_sizes={offset_dims}, collapsed_slice_dims="
                f"{collapsed_slice_dims}."
            )

        for i in range(len(slice_sizes)):
            slice_size = slice_sizes[i]
            corresponding_input_size = operand.shape[i]

            if not (
                core.greater_equal_dim(slice_size, 0) and core.greater_equal_dim(corresponding_input_size, slice_size)
            ):
                raise TypeError(
                    f"Slice size at index {i} in gather op is out of range, "
                    f"must be within [0, {corresponding_input_size} + 1), "
                    f"got {slice_size}."
                )

        for i in range(len(collapsed_slice_dims)):
            bound = slice_sizes[collapsed_slice_dims[i]]
            if bound != 1:
                raise TypeError(
                    f"Gather op can only collapse slice dims with bound 1, "
                    f"but bound is {bound} for index "
                    f"{collapsed_slice_dims[i]} at position {i}."
                )

        expanded_indices_shape.pop(index_vector_dim)
        indices_shape = iter(expanded_indices_shape)

        slice_sizes = (s for i, s in enumerate(slice_sizes) if i not in collapsed_slice_dims)
        return tuple(next(slice_sizes) if i in offset_dims else next(indices_shape) for i in range(output_shape_rank))

    @staticmethod
    def T(
        cts,
        operand,
        indices,
        *,
        dimension_numbers,
        slice_sizes,
        unique_indices,
        indices_are_sorted,
        mode,
        fill_value,
    ):
        assert ad.is_undefined_primal(operand)
        operand_shape = operand.symtensor.shape
        if type(t) is ad_util.Zero:
            out = ad_util.Zero(operand.symtensor)
        else:
            zeros = lax.full(operand_shape, lax._zero(t))
            scatter_dnums = ScatterDimensionNumbers(
                update_window_dims=dimension_numbers.offset_dims,
                inserted_window_dims=dimension_numbers.collapsed_slice_dims,
                scatter_dims_to_operand_dims=dimension_numbers.start_index_map,
            )
            out = scatter_add(
                zeros,
                indices,
                t,
                scatter_dnums,
                unique_indices=unique_indices,
                indices_are_sorted=indices_are_sorted,
                mode=mode,
            )
        return [out, None]


class Scatter(ShapeOp):
    get_impl = lambda: slope.RT.compiler.ScatterImpl

    @staticmethod
    def run(x, idx, *, dim):
        return [x.gatter(idx)]

    @staticmethod
    def vmap(
        dim_size,
        vals_in,
        dims_in,
        # scatter_op,
        # batched_args,
        # batch_dims,
        *,
        update_jaxpr,
        update_consts,
        dimension_numbers,
        indices_are_sorted,
        unique_indices,
        mode,
    ):
        operand, indices, updates = batched_args
        operand_bdim, indices_bdim, updates_bdim = batch_dims
        del update_jaxpr, update_consts  # Unused.

        # move the operand batch dim to the front if it is not None, otherwise create
        # it at the front (so that we can scatter into it)
        size = next(x.shape[ax] for x, ax in zip(batched_args, batch_dims) if ax is not None)
        operand = batching.bdim_at_front(operand, operand_bdim, size)
        operand_bdim = 0

        updates = batching.bdim_at_front(updates, updates_bdim, size)

        if indices_bdim is None:
            inserted_window_dims = tuple(np.add(1, dimension_numbers.inserted_window_dims))
            update_window_dims = (0,) + tuple(np.add(1, dimension_numbers.update_window_dims))
            scatter_dims_to_operand_dims = tuple(np.add(1, dimension_numbers.scatter_dims_to_operand_dims))
            dnums = ScatterDimensionNumbers(
                update_window_dims=update_window_dims,
                inserted_window_dims=inserted_window_dims,
                scatter_dims_to_operand_dims=scatter_dims_to_operand_dims,
            )
            return (
                scatter_op(
                    operand,
                    indices,
                    updates,
                    dnums,
                    indices_are_sorted=indices_are_sorted,
                    unique_indices=unique_indices,
                    mode=mode,
                ),
                0,
            )

        # see the third case in _gather_batching_rule for comparison and comments
        indices = batching.bdim_at_front(indices, indices_bdim, size)

        count_shape = list(indices.shape)
        count_shape[-1] = 1
        counts = lax.broadcasted_iota(indices.dtype, tuple(count_shape), 0)
        indices = lax.cat([counts, indices], len(count_shape) - 1)

        update_window_dims = tuple(np.add(1, dimension_numbers.update_window_dims))
        inserted_window_dims = (0,) + tuple(np.add(1, dimension_numbers.inserted_window_dims))
        scatter_dims_to_operand_dims = (0,) + tuple(np.add(1, dimension_numbers.scatter_dims_to_operand_dims))

        dnums = ScatterDimensionNumbers(
            update_window_dims=update_window_dims,
            inserted_window_dims=inserted_window_dims,
            scatter_dims_to_operand_dims=scatter_dims_to_operand_dims,
        )
        return (
            scatter_op(
                operand,
                indices,
                updates,
                dnums,
                indices_are_sorted=indices_are_sorted,
                unique_indices=unique_indices,
                mode=mode,
            ),
            0,
        )

    @staticmethod
    def jvp(
        primals,
        tangents,
        *,
        update_jaxpr,
        update_consts,
        dimension_numbers,
        indices_are_sorted,
        unique_indices,
        mode,
    ):
        operand, indices, updates = primals
        g_operand, g_indices, g_updates = tangents
        del g_indices  # ignored
        val_out = scatter_add_p.bind(
            operand,
            indices,
            updates,
            update_jaxpr=update_jaxpr,
            update_consts=update_consts,
            dimension_numbers=dimension_numbers,
            indices_are_sorted=indices_are_sorted,
            unique_indices=unique_indices,
            mode=mode,
        )
        if type(g_operand) is ad_util.Zero and type(g_updates) is ad_util.Zero:
            tangent_out = ad_util.Zero.from_value(val_out)
        else:
            g_operand = ad.instantiate_zeros(g_operand)
            g_updates = ad.instantiate_zeros(g_updates)
            tangent_out = scatter_add_p.bind(
                g_operand,
                indices,
                g_updates,
                update_jaxpr=update_jaxpr,
                update_consts=update_consts,
                dimension_numbers=dimension_numbers,
                indices_are_sorted=indices_are_sorted,
                unique_indices=unique_indices,
                mode=mode,
            )
        return val_out, tangent_out

    @staticmethod
    def typecheck(x: SymbolicTensor, idx, *, dim: Sequence[int]) -> List[SymbolicTensor]:
        shape = [x.shape[i] for i in dim]
        return [SymbolicTensor(shape, x.dtype)]

    @staticmethod
    def T(
        cts,
        operand,
        indices,
        updates,
        *,
        update_jaxpr,
        update_consts,
        dimension_numbers,
        indices_are_sorted,
        unique_indices,
        mode,
    ):
        assert not ad.is_undefined_primal(indices)
        if ad.is_undefined_primal(updates):
            updates_shape = updates.symtensor.shape
        else:
            updates_shape = updates.shape
        if type(t) is ad_util.Zero:
            operand_t = ad_util.Zero(operand.symtensor) if ad.is_undefined_primal(operand) else None
            update_t = ad_util.Zero(updates.symtensor) if ad.is_undefined_primal(updates) else None
        else:
            operand_t = update_t = None
            if ad.is_undefined_primal(operand):
                operand_t = t

            if ad.is_undefined_primal(updates):
                gather_dnums = GatherDimensionNumbers(
                    offset_dims=dimension_numbers.update_window_dims,
                    collapsed_slice_dims=dimension_numbers.inserted_window_dims,
                    start_index_map=dimension_numbers.scatter_dims_to_operand_dims,
                )
                slice_sizes = []
                pos = 0
                for i in range(len(t.shape)):
                    if i in dimension_numbers.inserted_window_dims:
                        slice_sizes.append(1)
                    else:
                        slice_sizes.append(updates_shape[dimension_numbers.update_window_dims[pos]])
                        pos += 1
                update_t = gather(
                    t,
                    indices,
                    dimension_numbers=gather_dnums,
                    slice_sizes=slice_sizes,
                    mode=mode,
                    fill_value=0,
                )
        return [operand_t, None, update_t]


@numpy_compiler.set_impl("scatter")
def fn(
    inputs,
    scatter_indices,
    updates,
    *,
    update_window_dims,
    inserted_window_dims,
    scatter_dims_to_operand_dims,
    index_vector_dim: int,
    slice_sizes,
    result_type,
    ret,
):
    return """
# SmallVector<Tensor> runScatterOp(
#     TensorRef<Tensor> inputs, const Tensor &scatterIndices,
#     TensorRef<Tensor> updates, const dim &updateWindowDims,
#     const dim &insertedWindowDims, const dim &scatterDimsToOperandDims,
#     dim indexVectorDim, Region &updateComputation, Scope &scope,
#     TensorRef<ShapedType> resultTypes) {
#   SmallVector<Tensor> results;
#   for (auto input : inputs) results.push_back(input);

#   dim updateScatterDims;
#   for (auto d : updates[0].getdim())
#     if (!llvm::is_contained(updateWindowDims, d))
#       updateScatterDims.push_back(d);

#   for (auto updateIndexIt = updates[0].index_begin();
#        updateIndexIt != updates[0].index_end(); ++updateIndexIt) {
#     auto updateIndex = *updateIndexIt;
#     Index updateScatterIndex;
#     for (auto d : updateScatterDims)
#       updateScatterIndex.push_back(updateIndex[d]);

#     auto startIndicesIndex = updateScatterIndex;
#     if (indexVectorDim < scatterIndices.getRank())
#       startIndicesIndex.insert(startIndicesIndex.begin() + indexVectorDim,
#                                kColon);
#     auto startIndex = runIndex(runSliceOp(scatterIndices, startIndicesIndex));

#     Index fullStartIndex(inputs[0].getRank(), 0);
#     for (auto dInput : inputs[0].getdim()) {
#       auto dStartIt = llvm::find(scatterDimsToOperandDims, dInput);
#       if (dStartIt == scatterDimsToOperandDims.end()) continue;
#       auto dStart = dStartIt - scatterDimsToOperandDims.begin();
#       fullStartIndex[dInput] = startIndex[dStart];
#     }

#     Index updateWindowIndex;
#     for (auto d : updateWindowDims) updateWindowIndex.push_back(updateIndex[d]);

#     Index fullWindowIndex(updateWindowIndex.size() + insertedWindowDims.size(),
#                           0);
#     for (size_t i = 0, wi = 0; i < fullWindowIndex.size(); ++i) {
#       if (llvm::is_contained(insertedWindowDims, i)) continue;
#       fullWindowIndex[i] = updateWindowIndex[wi++];
#     }

#     auto resultIndex = fullStartIndex + fullWindowIndex;
#     if (!resultIndex.inBounds(results[0].getShape())) continue;

#     SmallVector<InterpreterValue> updateComputationArgs;
#     for (auto result : results)
#       updateComputationArgs.push_back(
#           Tensor(RankedTensorType::get({}, result.getElementType()),
#                  result.get(resultIndex)));
#     for (auto update : updates)
#       updateComputationArgs.push_back(
#           Tensor(RankedTensorType::get({}, update.getElementType()),
#                  update.get(updateIndex)));

#     auto updatedValues = run(updateComputation, updateComputationArgs, &scope);
#     for (auto [result, updatedValue] : llvm::zip(results, updatedValues))
#       result.set(resultIndex, updatedValue.getTensor().get({}));
#   }

#   return results;
# }
"""


@numpy_compiler.set_impl("gather")
def fn(
    operand,
    start_indices,
    *,
    collapsed_slice_dims,
    start_index_map,
    offset_dims,
    index_vector_dim: int,
    slice_sizes,
):
    expanded_indices_shape = list(start_indices.shape)
    if len(expanded_indices_shape) == index_vector_dim:
        expanded_indices_shape.append(1)

    output_shape_rank = len(offset_dims) + start_indices.ndim - 1

    expanded_indices_shape.pop(index_vector_dim)
    indices_shape = iter(expanded_indices_shape)

    slice_sizes = (s for i, s in enumerate(slice_sizes) if i not in collapsed_slice_dims)
    res_size = tuple(next(slice_sizes) if i in offset_dims else next(indices_shape) for i in range(output_shape_rank))

    res = np.zeros(res_size)
    batch_dims = [d for d in list(range(res.ndim)) if d in offset_dims]

    for res_idx, _ in np.ndenumerate(res):
        batch_idx = [res_idx[d] for d in batch_dims]

    start_indices_idx = batch_idx[:]
    if index_vector_dim < start_indices.ndim:
        start_indices_idx.insert(index_vector_dim, -1)
    start_idx = start_indices[start_indices_idx]

    full_start_idx = [None] * operand.ndim
    for d in range(operand.ndim):
        dStartIt = start_index_map[d]
        if dStartIt == start_index_map[-1]:
            continue
        dStart = dStartIt - start_index_map[0]
        full_start_idx[d] = np.clip(start_idx[d], operand.shape[d] - slice_sizes[d])

    offset_idx = [res_idx[d] for d in offset_dims]
    full_offset_idx = [None] * (len(offset_dims) + len(collapsed_slice_dims))
    oi = 0
    for i in range(len(full_offset_idx)):
        if i in collapsed_slice_dims:
            continue
        full_offset_idx[i] = offset_idx[oi]
        oi += 1
    operand_idx = full_start_idx + full_offset_idx
    res[res_idx] = operand[operand_idx]
    return res



import slope


x = slope.ones((1, 3, 4, 4))
x_dot = slope.ones((1, 3, 4, 4))
w = slope.full((3, 8, 2, 2), 2.0)
w_dot = slope.full((3, 8, 2, 2), 2.0)
print(x.shape, w.shape)


def f(x, w):
    y = x.conv_transpose(w, stride=2)
    return y


y = f(x, w)
y, w_dot = slope.jvp(f, (x, w), (x_dot, w_dot))
print(f"{y.shape=}, {w.shape=}")
y, (gL_x, gL_w) = slope.value_and_grad(lambda *args: f(*args).sum())(x, w)
print(f"{y.shape=},  {gL_x.shape=}, {gL_w.shape=}")



import slope
import slope.nn as nn


class Block(nn.Module):
    def __init__(self, in_dim, out_dim, stride=1):
        self.conv1 = nn.Conv2d(in_dim, out_dim, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm(out_dim)

    def __call__(self, x, training=False):
        x = self.conv1(x)
        x = self.bn1(x, training)
        x = x.relu()
        return x


class Net(nn.Module):
    def __init__(self):
        self.block1 = Block(3, 8)
        self.flatten_fn = nn.Fn(lambda x: x.reshape((x.shape[0], -1)))
        self.linear = nn.Linear(8192, 10)

    def __call__(self, x, training=False):
        x = self.block1(x, training)
        x = self.flatten_fn(x)
        x = self.linear(x)
        return (x, self) if training else x


def loss_fn(model, batch):
    inputs, targets = batch
    preds, model_ = model(inputs, training=True)
    return -(preds.log_softmax() * targets).sum(), model_


g_loss_fn = slope.value_and_grad(loss_fn, has_aux=True)


@slope.jit
def gstep(model, batch):
    (loss, model_), g_model = g_loss_fn(model, batch)
    return loss, g_model, model_


model = Net()
_, treedef = slope.core.tree_flatten(model)
print(treedef)
x = slope.rand((1, 3, 32, 32))
y = slope.ones((1,)).one_hot(10, dtype=slope.float32)
loss, g_model, model_ = gstep(model, (x, y))



import slope

x = slope.ones(5)
w = slope.arange(5, dtype=x.dtype)
y_eq = (x == w).int()
y_ne = (x != w).int()
y_lt = (x < w).int()
y_le = (x <= w).int()
y_ge = (x >= w).int()
y_gt = (x > w).int()


print(f"{x=}")
print(f"{w=}")
print(f"x == w = {y_eq}")
print(f"x != w = {y_ne}")
print(f"x <  w = {y_lt}")
print(f"x <= w = {y_le}")
print(f"x >  w = {y_gt}")
print(f"x >= w = {y_ge}")



import math
import slope
from slope import nn
from functools import partial


def conv3x3(in_planes, out_planes, stride=1):
    "3x3 convolution with padding"
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, cfg, stride=1, downsample=None):
        # cfg should be a number in this case
        self.conv1 = conv3x3(inplanes, cfg, stride)
        self.bn1 = nn.BatchNorm2d(cfg)
        self.relu = nn.ReLU()
        self.conv2 = conv3x3(cfg, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def __call__(self, x, training=False):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out, training)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out, training)

        if self.downsample is not None:
            residual = self.downsample(x)

        out = out + residual
        out = self.relu(out)

        return out


def downsample_basic_block(x, planes):
    x = slope.avgpool2d(x, (2, 2))
    zero_pads = slope.zeros(x.size(0), planes - x.size(1), x.size(2), x.size(3))
    out = slope.cat([x, zero_pads], dim=1)
    return out


class ResNet(nn.Module):
    def __init__(self, depth, dataset="cifar10", cfg=None):
        # Model type specifies number of layers for CIFAR-10 model
        assert (depth - 2) % 6 == 0, "depth should be 6n+2"
        n = (depth - 2) // 6

        block = BasicBlock
        if cfg == None:
            cfg = [[16] * n, [32] * n, [64] * n]
            cfg = [item for sub_list in cfg for item in sub_list]

        self.cfg = cfg

        self.inplanes = 16
        self.conv1 = nn.Conv2d(3, 16, kernel_size=3, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.relu = nn.ReLU()
        self.layer1 = self._make_layer(block, 16, n, cfg=cfg[0:n])
        self.layer2 = self._make_layer(block, 32, n, cfg=cfg[n : 2 * n], stride=2)
        self.layer3 = self._make_layer(block, 64, n, cfg=cfg[2 * n : 3 * n], stride=2)
        self.avgpool = nn.AvgPool2d(8)
        if dataset == "cifar10":
            num_classes = 10
        elif dataset == "cifar100":
            num_classes = 100
        self.fc = nn.Linear(64 * block.expansion, num_classes)

    def _make_layer(self, block, planes, blocks, cfg, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = partial(downsample_basic_block, planes=planes * block.expansion)

        layers = []
        layers.append(block(self.inplanes, planes, cfg[0], stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, cfg[i]))

        return nn.Sequential(*layers)

    def __call__(self, x, training=False):
        x = self.conv1(x)
        x = self.bn1(x, training)
        x = self.relu(x)  # 32x32

        x = self.layer1(x, training)  # 32x32
        x = self.layer2(x, training)  # 16x16
        x = self.layer3(x, training)  # 8x8

        x = self.avgpool(x)
        x = x.reshape(x.size(0), -1)
        x = self.fc(x)

        return (x, self) if training else x


def resnet(**kwargs):
    return ResNet(**kwargs)


if __name__ == "__main__":
    net = resnet(depth=20)
    x = slope.tensor.rand(16, 3, 32, 32)
    y = net(x)
    print(y.data.shape)



import slope
import math

XDIMS = (3, 4, 5)

x = slope.arange(math.prod(XDIMS), dtype=slope.float32).reshape(*XDIMS)
x_dot = x.ones_like()
print(f"{x.shape=}")


def f(x):
    return x.sum(-1)


y = f(x)
y_vmap = slope.vmap(f)(x)
y_jit_vmap = slope.jit(slope.vmap(f))(x)
y_vmap_jit = slope.vmap(slope.jit(f))(x)
print(f"{y.shape=}, {y_vmap=} {y_jit_vmap.shape=},  {y_vmap_jit.shape=}")
y_jvp, y_dot_jvp = slope.jvp(f, (x,), (x_dot,))
y_jit_jvp, y_dot_jit_jvp = slope.jvp(slope.jit(f), (x,), (x_dot,))
print(f"{y_jvp.shape=}")
print(f"{y_jit_jvp.shape=}, {y_dot_jit_jvp.shape=}")

loss_fn = lambda *args: f(*args).sum()
(
    L,
    gL_x,
) = slope.value_and_grad(
    loss_fn, argnums=(0,)
)(x)
(
    L_jit,
    gL_x_jit,
) = slope.value_and_grad(
    slope.jit(loss_fn), argnums=(0,)
)(x)
print(f"{L.shape=}, {gL_x.shape=}")
print(f"{L_jit.shape=}, {gL_x_jit.shape=}")



import slope
from slope.core import (
    Compiler,
    Backend,
    Operator,
    OperatorSet,
    ProcedureSet,
    Tensor,
    TensorBuffer,
    SymbolicTensor,
    UndefinedPrimal,
    list_zip,
    list_map,
)

import math
import numpy as np
from typing import (
    Tuple,
    List,
    Dict,
    Any,
    Optional,
    Sequence,
    Union,
    Iterator,
    Callable,
)
from collections import defaultdict
import importlib
import os
import functools

# --------------
# Compiler
# --------------

compiler = Compiler(name="numpy", default_dtype=slope.SLOPE_DTYPE)
compiler.set_dtype_map(
    {
        slope.core.dtypes.float32: np.dtype("float32"),
        Tensor.int64: np.dtype("int64"),
        Tensor.int32: np.dtype("int32"),
        Tensor.int8: np.dtype("int8"),
        Tensor.bool: np.dtype("bool"),
    }
)


@compiler.set_method
def from_numpy(self, val, dtype=compiler.default_dtype_value):
    val = np.array(val, dtype=compiler.dtype_map[dtype])
    return Tensor(TensorBuffer(val))


@compiler.set_method
def numpy_of(self, tensor):
    return tensor.buf.val


@compiler.set_method
def device_of(self, tensor):
    return "cpu"


@compiler.set_method
def shape_of(self, tensor):
    return tuple(int(i) for i in tensor.buf.val.shape)


@compiler.set_method
def dtype_of(self, tensor):
    return self.dtype_map_inv[tensor.buf.val.dtype]


@compiler.set_method
def export(self, jit_output: slope.core.JitOutput, output_path, *args, **kwargs):
    code = jit_output.code
    os.makedirs(output_path, exist_ok=True)
    consts_dir_path = os.path.join(output_path, "consts")
    os.makedirs(consts_dir_path, exist_ok=True)
    in_binders = jit_output.codegen_output["in_binders"]
    outs = jit_output.codegen_output["outs"]
    num_consts = jit_output.program.num_consts
    load_consts_code = ""
    for i in range(num_consts):
        const_name = in_binders[i]["name"]
        const_path = os.path.join(consts_dir_path, f"{const_name}.npy")
        load_consts_code += f"""{const_name} = np.load(os.path.join(consts_dir_path, "{const_name}.npy"))\n"""
        np.save(const_path, in_binders[i]["type"].numpy())
    input_args_code = ", ".join(ib["name"] for ib in in_binders[num_consts:])
    args_code = ", ".join(ib["name"] for ib in in_binders)
    input_arg_names = [ib["name"] for ib in in_binders[num_consts:]]
    input_arg_names_str = ", ".join(input_arg_names)
    outs_names = [out["name"] for out in outs]

    test_input_code = ""
    for i in range(num_consts, len(in_binders)):
        input_name = in_binders[i]["name"]
        input_shape = in_binders[i]["type"].shape
        dtype = in_binders[i]["type"].dtype
        input_dtype = ("np." + dtype.numpy.__name__) if dtype is not Tensor.bool else "bool"
        test_input_code += f"""    {input_name} = np.ones({input_shape}, dtype={input_dtype})\n"""

    module_path = os.path.join(output_path, "__init__.py")
    module_code = f"""import numpy as np
import os
root_path = os.path.dirname(__file__)
consts_dir_path =  os.path.join(root_path, "consts")
{load_consts_code}
{code}

input_arg_names = {input_arg_names}
out_names = {outs_names}

def run({input_args_code}):
    return main({args_code})

if __name__ == "__main__":
{test_input_code}
    for inp_name, inp in zip(input_arg_names, ({input_arg_names_str})):
        print(f"{{inp_name}} = ")
        print(inp)
        print(f"dtype: {{inp.dtype}}")
        print(f"shape: {{inp.shape}}")
        print()

    outs = run({input_arg_names_str})

    print("outputs:")
    for out_name, out in zip(out_names, outs):
        print(f"{{out_name}} = ")
        print(out)
        print(f"dtype: {{out.dtype}}")
        print(f"shape: {{out.shape}}")
        print()
"""
    with open(module_path, "w") as f:
        f.write(module_code)
        slope.dblog(module_code, enable=slope.LOG_JIT)


@compiler.set_method
def compile(self, codegen_output):
    deps_dict = dict()
    deps_dict["numpy"] = importlib.import_module("numpy")
    deps_dict["np"] = deps_dict["numpy"]
    deps_dict["math"] = importlib.import_module("math")
    code_lines = codegen_output["code_lines"]
    exec_locals = dict()
    code = "\n".join(code_lines)
    exec(compile(code, "<string>", "exec"), deps_dict, exec_locals)
    fn = exec_locals["main"]
    return fn, code


@compiler.set_method
def codegen(self, program, args, *, fn_name: str = "main", fn_defs=dict()) -> List[Any]:
    if fn_name == "main":
        assert not hasattr(self, "fn_count")
        self.fn_count = 0
        assert not hasattr(self, "depth")
        self.depth = 0

    def indent(code, amount):
        spaces = " " * (len(code) - len(code.lstrip()))
        spaces += " " * amount
        return "\n".join([spaces + line for line in code.strip().split("\n")])

    # codegen is recursive if jit-of-jit happens
    backend: Dict[slope.Var, Any] = {}
    il1 = (self.depth + 1) * 4
    body_code_lines = []

    for inb in program.in_binders:
        prefix = "x" if type(inb.symtensor) is SymbolicTensor else "c"
        idx = sum_py([1 if v["name"][0] == prefix else 0 for v in backend.values()])
        backend[inb] = dict(name=f"{prefix}{idx}", type=inb.symtensor)

    for instruction in program.instructions:
        if len(instruction.out_binders) == 0:  # skip codegen for function returns nothing
            continue
        in_vals = list_map(lambda x: backend[x]["name"], instruction.inputs)
        for outb in instruction.out_binders:
            prefix = "y" if outb in program.outs else "z"
            idx = sum_py([1 if v["name"][0] == prefix else 0 for v in backend.values()])
            backend[outb] = dict(name=f"{prefix}{idx}", type=outb.symtensor)

        out_vals = list_map(lambda z: backend[z]["name"], instruction.out_binders)
        if instruction.op.op_type is slope.core.OperatorType.Meta:
            lhs = ", ".join(out_vals)
            self.depth += 1
            rhs, fn_defs = self.impls[instruction.op](program, args, instruction, in_vals, fn_defs)
            self.depth -= 1
            impl_code = f"{lhs} = {rhs}"
        else:
            impl_code = self.impls[instruction.op](*in_vals, **instruction.params)
            if len(out_vals) == 1:
                impl_code = impl_code.replace("ret", out_vals[0])
            else:
                raise NotImplementedError
        for np_dtype in self.dtype_map.values():  # fix dtype kwargs not having 'np.' prefix
            impl_code = impl_code.replace(
                np_dtype.name,
                "bool" if np_dtype is np.dtype("bool") else f"np.{np_dtype.name}",
            )

        for impl_code_line in impl_code.split("\n"):  # handle multi-line code
            body_code_lines += [indent(impl_code_line, il1)]

    in_binders = list_map(lambda x: backend[x], program.in_binders)
    arg_type_strs = [i["name"] for i in in_binders]
    # arg_type_asserts = [
    #     f"{self.dtype_map[i['type'].dtype]}[{repr(list(i['type'].shape))[1:-1]}] {i['name']}" for i in in_binders
    # ]
    fn_args_str = ", ".join(arg_type_strs)

    outs = list_map(lambda x: backend[x], program.outs)  # TODO: input that is output should has identity op
    out_type_strs = [o["name"] for o in outs]
    # out_type_asserts = [
    #     f"{self.dtype_map[o['type'].dtype]}[{repr(list(o['type'].shape))[1:-1]}] {o['name']}" for o in outs
    # ]

    head_code_lines = []
    head_code_lines += [f"def {fn_name} ({fn_args_str}):"]
    out_type_str = ", ".join(out_type_strs) + ("," if len(outs) == 1 else "")
    return_line = [indent(f"return {out_type_str}", il1)]

    functions_code_lines = []
    for op, fn_def_code_lines in fn_defs.items():
        # functions_code_lines += fn_def_code_lines
        functions_code_lines += fn_def_code_lines

    code_lines = head_code_lines + [indent(l, il1) for l in functions_code_lines] + body_code_lines + return_line
    slope.dblog(
        f"\n-- {program.name} codegen:\n\n" + "\n".join(code_lines) + "\n\n==\n",
        enable=slope.LOG_JIT,
    )

    if fn_name == "main":
        del self.fn_count
        del self.depth
    assert len(outs) == len(program.outs)
    return dict(code_lines=code_lines, fn_defs=fn_defs, in_binders=in_binders, outs=outs)


### Operator Impls

compiler.set_impl(operator_set.cast)(lambda self, x, *, dtype: f"ret = {x}.astype(dtype={dtype})")
compiler.set_impl(operator_set.stop_gradient)(lambda self, x: f"ret = {x}")
compiler.set_impl(operator_set.neg)(lambda self, x: f"ret = np.negative({x})")
compiler.set_impl(operator_set.sqrt)(lambda self, x: f"ret = np.sqrt({x})")
compiler.set_impl(operator_set.exp)(lambda self, x: f"ret = np.exp({x})")
compiler.set_impl(operator_set.log)(lambda self, x: f"ret = np.log({x})")
compiler.set_impl(operator_set.sin)(lambda self, x: f"ret = np.sin({x})")
compiler.set_impl(operator_set.add)(lambda self, x, w: f"ret = np.add({x}, {w})")
compiler.set_impl(operator_set.sub)(lambda self, x, w: f"ret = np.subtract({x}, {w})")
compiler.set_impl(operator_set.mul)(lambda self, x, w: f"ret = np.multiply({x}, {w})")
compiler.set_impl(operator_set.div)(lambda self, x, w: f"ret = np.divide({x}, {w})")
compiler.set_impl(operator_set.pow)(lambda self, x, w: f"ret = np.power({x}, {w})")
compiler.set_impl(operator_set.invert)(lambda self, x: f"ret = np.invert({x})")
compiler.set_impl(operator_set.equal)(lambda self, x, w: f"ret = np.equal({x}, {w})")
compiler.set_impl(operator_set.maximum)(lambda self, x, w: f"ret = np.maximum({x}, {w})")
compiler.set_impl(operator_set.sum)(lambda self, x, *, dim, keepdim: f"ret = np.sum({x}, axis={dim}, keepdims={keepdim})")
compiler.set_impl(operator_set.max)(lambda self, x, *, dim, keepdim: f"ret = np.max({x}, axis={dim}, keepdims={keepdim})")
compiler.set_impl(operator_set.arange)(
    lambda self, *, start, stop, stride, dtype: f"ret = np.arange(start={start}, stop={stop}, step={stride}, dtype={dtype})"
)
compiler.set_impl(operator_set.full)(
    lambda self, *, shape, fill_value, dtype: f"ret = np.full(shape={shape}, fill_value={fill_value}, dtype={dtype})"
)

compiler.set_impl(operator_set.random_uniform)(
    lambda self, *, shape, dtype: (
        f"ret = {'np.array(' if shape == () else ''}np.random.uniform(low=np.zeros(shape={shape})){')' if shape == () else ''}.astype(dtype={dtype})"
    )
)
compiler.set_impl(operator_set.random_normal)(
    lambda self, *, shape, dtype: (
        f"ret = {'np.array(' if shape == () else ''}np.random.normal(loc=np.zeros(shape={shape})){')' if shape == () else ''}.astype(dtype={dtype})"
    )
)
compiler.set_impl(operator_set.expand)(lambda self, x, *, shape: f"ret = np.broadcast_to({x}, shape={shape})")

compiler.set_impl(operator_set.reshape)(lambda self, x, *, shape: f"ret = np.reshape({x}, newshape={shape})")


@compiler.set_impl(operator_set.pad)
def pad_impl(self, x, *, padding, mode, value):
    pad_width = [(lo, hi) for lo, hi in zip(padding[0::2], padding[1::2])]
    return f"ret = np.pad({x}, {pad_width}, constant_values={value})"


compiler.set_impl(operator_set.slice)(
    lambda self, x, *, starts, limits, strides: f"ret = {x}[tuple(slice(s, l, st) for s, l, st in zip({starts}, {limits}, {strides}))]"
)

compiler.set_impl(operator_set.cat)(lambda self, *xs, dim: f"ret = np.concatenate(({','.join(xs)}), axis={dim})")
compiler.set_impl(operator_set.permute)(lambda self, x, *, perm: f"ret = np.transpose({x}, axes={perm})")
compiler.set_impl(operator_set.flip)(lambda self, x, *, dim: f"ret = np.flip({x}, axis={dim})")


@compiler.set_impl(slope.core.jit_op)
def jit_op_impl(self, program, args, instruction, in_vals, fn_defs):
    jit_program = instruction.params["program"]
    jit_name = f"{program.name}"
    jit_codegen_output = self.codegen(
        jit_program,
        args,
        fn_name=jit_name,
        fn_defs=fn_defs,
    )
    assert jit_name not in fn_defs.keys()
    fn_defs[jit_name] = jit_codegen_output["code_lines"]
    fn_defs = {**fn_defs, **jit_codegen_output["fn_defs"]}
    args_str = ", ".join(in_vals)
    rhs = f"{jit_name}({args_str})"
    return rhs, fn_defs


procedure_set = ProcedureSet()


@procedure_set.register()
def zeros(*args, **kwargs):
    dtype = kwargs.get("dtype", slope.SLOPE_DTYPE)
    if kwargs.get("shape", None) is None:
        shape = args[0] if isinstance(args[0], (tuple, list)) else args
        assert all(i >= 0 for i in shape)
    return slope.full(shape, 0.0, dtype)


@procedure_set.register()
def ones(*args, **kwargs):
    dtype = kwargs.get("dtype", slope.SLOPE_DTYPE)
    if kwargs.get("shape", None) is None:
        shape = args[0] if isinstance(args[0], (tuple, list)) else args
        assert all(i >= 0 for i in shape)
    return slope.full(shape=shape, fill_value=1.0, dtype=dtype)


@procedure_set.register(static_argnames="fill_value")
def full_like(y, fill_value):
    return slope.full(shape=y.shape, fill_value=fill_value, dtype=y.dtype)


@procedure_set.register()
def zeros_like(y):
    return full_like(y, 0.0)


@procedure_set.register()
def ones_like(y):
    return full_like(y, 1.0)


@procedure_set.register()
def where(x, trueval, falseval):
    cond = x != 0.0
    if not isinstance(trueval, Tensor):
        trueval = slope.full((), trueval)
    if not isinstance(falseval, Tensor):
        falseval = slope.full((), falseval)
    cond = cond.cast(trueval.dtype)
    return cond * trueval + (1.0 - cond) * falseval


@procedure_set.register(static_argnames="dim keepdim")
def mean(x, dim=None, keepdim=False):
    out = x.sum(dim=dim, keepdim=keepdim)
    return out * (math.prod(out.shape) / math.prod(x.shape))


@procedure_set.register()
def rsqrt(x):
    return (slope.ones_like(x) / x).sqrt()


@procedure_set.register()
def cos(x):
    return ((math.pi / 2) - x).sin()


@procedure_set.register()
def tan(x):
    return x.sin() / x.cos()


@procedure_set.register()
def not_equal(x, w):
    return ~(x.equal(w))


@procedure_set.register()
def greater_equal(x, w):
    return x.maximum(w).equal(w)


@procedure_set.register()
def less_equal(x, w):
    return x.minimum(w).equal(w)


@procedure_set.register()
def greater(x, w):
    return 1.0 - (x <= w)


@procedure_set.register()
def less(x, w):
    return 1.0 - (x >= w)


@procedure_set.register()
def minimum(x, w):
    return -x.maximum(-x, -w)


@procedure_set.register(static_argnames="dim keepdim")
def min(x, dim=None, keepdim=False):
    return -((-x).max(x, dim, keepdim))


@procedure_set.register(static_argnames="dim keepdim")
def argmax(x, dim=None, keepdim=False):
    if dim is None:
        idx = (x == x.max(dim)) * slope.arange(
            math.prod(x.shape) - 1,
            -1,
            -1,
            dtype=slope.int32,
        ).reshape(x.shape)
        return math.prod(x.shape) - idx.max() - 1
    dim = dim + len(x.shape) if dim < 0 else dim
    m = (x == x.max(dim=dim, keepdim=True)).cast(slope.int32)
    idx = m * slope.arange(x.shape[dim] - 1, -1, -1, dtype=slope.int32).reshape((x.shape[dim], *[1] * (x.ndim - dim - 1)))
    ret = x.shape[dim] - idx.max(dim=dim, keepdim=keepdim) - 1
    return ret


@procedure_set.register(static_argnames="dim keepdim")
def argmin(x, dim=None, keepdim=False):
    return (-x).argmax(dim=dim, keepdim=keepdim)


@procedure_set.register()
def log2(x):
    return x.log() / math.log(2)


@procedure_set.register()
@staticmethod
def _tri(r: int, c: int, k: int = 0, **kwargs) -> Tensor:
    return slope.arange(r, **kwargs).unsqueeze(1).expand(r, c) <= Tensor.arange(-k, c - k, **kwargs).unsqueeze(0).expand(r, c)


@procedure_set.register()
def triu(self, k: int = 0) -> Tensor:
    return slope._tri(
        self.shape[-2],
        self.shape[-1],
        k=k,
        dtype=self.dtype,
        device=self.device,
    ).where(self, slope.zeros_like(self))


@procedure_set.register()
def tril(self, k: int = 0) -> Tensor:
    return slope._tri(
        self.shape[-2],
        self.shape[-1],
        k=k + 1,
        dtype=self.dtype,
        device=self.device,
    ).where(slope.zeros_like(self), self)


@procedure_set.register()
def trunc(self: Tensor) -> Tensor:
    return self.cast(slope.int32).cast(self.dtype)


@procedure_set.register()
def ceil(self: Tensor) -> Tensor:
    return (self > (b := self.trunc())).where(b + 1, b)


@procedure_set.register()
def floor(self: Tensor) -> Tensor:
    return (self < (b := self.trunc())).where(b - 1, b)


@procedure_set.register()
def square(self):
    return self * self


@procedure_set.register()
def clip(self, min_, max_):
    return self.maximum(min_).minimum(max_)


@procedure_set.register()
def abs(self):
    return self.relu() + (-self).relu()


@procedure_set.register()
def sign(self):
    return self / (self.abs() + 1e-10)


@procedure_set.register()
def reciprocal(self):
    return 1.0 / self


@procedure_set.register()
def matmul(x, w):
    x = x.reshape((*x.shape[0:-1], 1, x.shape[-1]))
    w = w.reshape((*w.shape[0:-2], 1, w.shape[-2], w.shape[-1])).T()
    return (x * w).sum(-1).reshape((*x.shape[0:-2], -1))


@procedure_set.register()
def T(x):
    perm = list(range(x.ndim))
    perm[-2], perm[-1] = perm[-1], perm[-2]
    return x.permute(tuple(perm))


@procedure_set.register()
def getitem(self, val):
    # Union[int, slice, Tensor, None, Ellipsis, Tuple[Union[int, slice, Tensor, None, Ellipsis], ...]]
    def normalize_int(e, i, dim_sz):
        if -dim_sz <= e < dim_sz:
            return e if e != -1 else dim_sz - 1
        raise IndexError(f"index {e} is out of bounds for dimension {i} with size {self.shape[i]}")

    orig_slices = list(val) if isinstance(val, tuple) else [val]
    count = defaultdict(list)
    for i, v in enumerate(orig_slices):
        count[type(v) if not isinstance(v, slope.core.Tensor) else "tensor"] += [i]

    if (num_slices := len(count[int]) + len(count[slice_py]) + len(count["tensor"])) > len(self.shape):
        raise IndexError(f"too many indices for tensor of dimension {len(self.shape)}")
    if len(ellipsis_found := count[type(Ellipsis)]) > 1:
        raise IndexError("an index can only have a single ellipsis ('...')")

    ellipsis_idx = ellipsis_found[0] if ellipsis_found else len(orig_slices)
    orig_slices[ellipsis_idx : ellipsis_idx + 1] = [slice_py(None)] * (len(self.shape) - num_slices)

    valid_slices = [v for v in orig_slices if v is not None]
    valid_slices = [
        v if isinstance(v, slice_py) else slice_py(y_ := normalize_int(v, i, dim_sz), y_ + 1) if isinstance(v, int) else slice_py(None)
        for i, (v, dim_sz) in enumerate(zip(valid_slices, self.shape))
    ]

    start, stop, strides = zip(*y) if (y := [s.indices(dim_sz) for s, dim_sz in zip(valid_slices, self.shape)]) else ((), (), ())
    new_slice = tuple((s, e) if st > 0 else (e + 1, s + 1) for s, e, st in zip(start, stop, strides))
    sliced_tensor = self.padslice(new_slice).flip(dim=tuple([i for i, s in enumerate(strides) if s < 0]))
    new_shape = sliced_tensor.shape
    if any(abs_py(s) != 1 for s in strides):
        strides = tuple(abs_py(s) for s in strides)
        # Pad: add pad at the end: [dim_sz] -> [dim_sz_padded]
        padded_tensor = sliced_tensor.pad(
            tuple((0, s - (dim_sz % s) if dim_sz % s != 0 else 0) for s, dim_sz in zip(strides, sliced_tensor.shape))
        )
        # Reshape: [dim_sz_padded] -> [dim_sz_padded // s, s]
        reshaped_tensor = padded_tensor.reshape(flatten([sh // s, s] for sh, s in zip(padded_tensor.shape, strides)))
        new_shape = reshaped_tensor.shape[::2]
        # Shrink: do [:, 0]
        sliced_tensor = reshaped_tensor.padslice(tuple(flatten(((0, sh), (0, 1)) for sh in new_shape)))

    final_shape, it_shape, dim, tensors, dim_collapsed = (
        [],
        iter(new_shape),
        [],
        [],
        0,
    )
    for i, s in enumerate(orig_slices):
        if s is None:
            final_shape.append(1)
        else:  # s is int or slice or Tensor
            dim_shape = next(it_shape)
            if isinstance(s, int):
                dim_collapsed += 1
            else:
                final_shape.append(dim_shape)
                if isinstance(s, slope.core.Tensor):
                    tensors.append(s)
                    dim.append(i - dim_collapsed)
    ret = sliced_tensor.reshape(tuple(final_shape))

    if tensors:  # Fancy/tensor indexing
        # normalize idx
        idx = [t.sign().neg().relu() * ret.shape[d] + t for d, t in zip(dim, tensors)]
        max_dim = max(i.ndim for i in idx)
        # compute sum_dim, arange, and idx
        sum_dim = [d if n == 0 else d + max_dim - n for n, d in enumerate(dim)]
        slice_arange = [
            slope.arange(
                ret.shape[d],
                dtype=slope.int32,
                requires_grad=False,
                device=self.device,
            ).reshape(
                *[1] * sd,
                ret.shape[d],
                *[1] * (ret.ndim + max_dim - n - sd - 1),
            )
            for n, (sd, d) in enumerate(zip(sum_dim, dim))
        ]
        first_idx = [
            idx[0].reshape(
                *[1] * dim[0],
                *[1] * (1 + max_dim - idx[0].ndim),
                *idx[0].shape,
                *[1] * (ret.ndim - dim[0] - 1),
            )
        ]
        rest_idx = [
            i.reshape(
                *[1] * dim[0],
                *[1] * (max_dim - i.ndim),
                *i.shape,
                *[1] * (ret.ndim - dim[0] - n),
            )
            for n, i in enumerate(idx[1:], 1)
        ]
        idx = first_idx + rest_idx
        ret = ret.reshape(
            *ret.shape[: sum_dim[0] + 1],
            *[1] * max_dim,
            *ret.shape[sum_dim[0] + 1 :],
        )
        # iteratively fancy index
        for a, i, sd in zip(slice_arange, idx, sum_dim):
            ret = (a == i).mul(ret).sum(sd)
        # special permute case
        if dim[0] != 0 and len(dim) != 1 and dim != list(range(dim[0], dim[-1] + 1)):
            ret_dims = list(range(ret.ndim))
            ret = ret.permute(ret_dims[dim[0] : dim[0] + max_dim] + ret_dims[: dim[0]] + ret_dims[dim[0] + max_dim :])
    return ret


@procedure_set.register(static_argnames=("arg", "value"))
def padslice(x, arg: Sequence[Optional[Tuple[int, int]]], value: float = 0):
    def flatten_seq(l: Iterator):
        return [item for sublist in l for item in sublist]

    # some dim are pad, some are sliced
    arg_ = tuple([a if a is not None else (0, s) for s, a in zip(x.shape, arg)])
    padding = tuple([(max_py(0, -p[0]), max_py(0, p[1] - x.shape[i])) for i, p in enumerate(arg_)])
    x = x.pad(flatten_seq(padding), value=value)  # flatten
    starts, limits, strides = tuple(zip(*[(p[0] + padding[i][0], p[1] + padding[i][0], 1) for i, p in enumerate(arg_)]))
    x = x.slice(starts, limits, strides)
    return x


@procedure_set.register(static_argnames="padding value")
def pad2d(x, padding: Union[List[int], Tuple[int, ...]], value: float = 0):
    # (padding_left, padding_right, padding_top, padding_bottom)
    slc = [(-p0, s + p1) for p0, p1, s in zip(padding[::2], padding[1::2], x.shape[::-1])][::-1]
    return x.padslice([(0, s) for s in x.shape[: -(len(padding) // 2)]] + slc, value=value)


@procedure_set.register(static_argnames="dim")
def gather(x, idx, dim: int):
    assert idx.ndim == x.ndim, "x.ndim must equal idx.ndim"
    assert all(s >= i for s, i in zip(x.shape, idx.shape)), "all dim of idx.shape must be smaller than x.shape"
    if dim < 0:
        dim += x.ndim
    idx = idx.transpose(ax=dim, aw=0).expand_dims(-1)
    permarg = list(range(x.ndim))
    permarg = permarg[1:dim] + [permarg[0]] + permarg[dim + 1 :] + [permarg[dim]] if dim != 0 else permarg[1:] + [permarg[0]]
    return (
        (
            (
                idx
                == slope.arange(
                    x.shape[dim],
                    dtype=slope.int32,
                    requires_grad=False,
                    device=x.device,
                )
            )
            * x.permute(*permarg).padslice(tuple([*[(0, sh) for sh in idx.shape[1:-1]], (0, x.shape[dim])])).expand_dims(0)
        )
        .sum(-1)
        .transpose(ax=0, aw=dim)
    )


@procedure_set.register(static_argnames="dim")
@staticmethod
def stack(tensors, dim=0):
    first = tensors[0].expand_dims(dim)
    expand_dimsd_tensors = [tensor.expand_dims(dim) for tensor in tensors[1:]]
    return first.cat(*expand_dimsd_tensors, dim=dim)


@procedure_set.register(static_argnames="repeats")
def repeat(x, repeats):
    base_shape = (1,) * (len(repeats) - x.ndim) + x.shape
    new_shape = [x for b in base_shape for x in [1, b]]
    expand_shape = [x for rs in zip(repeats, base_shape) for x in rs]
    final_shape = [r * s for r, s in zip(repeats, base_shape)]
    return x.reshape(new_shape).broadcast(expand_shape).reshape(final_shape)


@procedure_set.register(static_argnames="dim")
def split(x, num: int, dim: int):
    dim, step = dim + x.ndim if dim < 0 else dim, math.ceil(x.shape[dim] / num)
    slice_params = [[slice(None)] * dim + [slice(k, k + step)] for k in range(0, x.shape[dim], step)]
    return tuple(x[tuple(sl)] for sl in slice_params)


@procedure_set.register(static_argnames="dim")
def squeeze(x, dim=None):
    if dim is None:
        return x if 1 not in x.shape else x.reshape(*[size for size in x.shape if size != 1])
    if dim <= 0 and x.ndim == 0:
        return x  # This is to match PyTorch behavior
    if not -x.ndim <= dim < x.ndim:
        raise IndexError(
            f"Dimension out of range (expected to be in range of [{-x.ndim if x.ndim > 0 else x.ndim-1}, {x.ndim-1 if x.ndim > 0 else x.ndim}], but got {dim})"
        )
    if dim < 0:
        dim += x.ndim
    return x if x.shape[dim] != 1 else x.reshape(*[size for idx, size in enumerate(x.shape) if idx != dim])


@procedure_set.register(static_argnames="dim")
def expand_dims(x, dim):
    if dim < 0:
        dim = len(x.shape) + dim + 1
    return x.reshape(x.shape[:dim] + (1,) + x.shape[dim:])


@procedure_set.register(static_argnames="ax aw")
def transpose(x, ax=1, aw=0):
    order = list(range(len(x.shape)))
    order[ax], order[aw] = order[aw], order[ax]
    return x.permute(tuple(order))


@procedure_set.register(static_argnames="start_dim")
def flatten(x, start_dim=0):
    return x.reshape(shape=x.shape[:start_dim] + (-1,))


@procedure_set.register(static_argnames="dim")
def cumsum(x, dim: int = 0):
    return x.transpose(dim, -1).pad((x.shape[dim] - 1, 0)).pool((x.shape[dim],)).sum(-1).transpose(dim, -1)


@staticmethod
@procedure_set.register(static_argnames="start stop step")
def arange_with_cumsum(start, stop=None, step=1):
    if stop is None:
        stop, start = start, 0
    return slope.full((math.ceil((stop - start) / step),), step).cumsum() + (start - step)


@procedure_set.register(static_argnames="dtype")
def one_hot(x, k, dtype=Tensor.int32):
    return (x[:, None].cast(dtype) == slope.arange(k, dtype=dtype)).cast(dtype)


numpy_backend = Backend(operator_set, procedure_set, compiler)



"""
# run with different backends
SLOPE_BACKEND=iree python examples/nn/mnist_mlp.py
SLOPE_BACKEND=onnxruntime python examples/nn/mnist_mlp.py

# print backend code
LOG_JIT=1 SLOPE_BACKEND=iree python examples/nn/mnist_mlp.py
"""

import slope
import slope.nn as nn

import time
import itertools
import math
import numpy as np
from tqdm import tqdm

from lib.datasets.mnist import get_mnist

import numpy as np


def loss_fn(model, batch):
    x, y_onehot = batch
    preds = model(x)
    return -(preds.log_softmax(-1) * y_onehot).sum()


value_and_gloss_fn = slope.value_and_grad(loss_fn)


@slope.jit
def train_step(model, batch, optimizer):
    loss, gloss_model = value_and_gloss_fn(model, batch)
    new_model, new_optimizer = optimizer(model, gloss_model)
    return loss, new_model, new_optimizer


@slope.jit
def test_all(model, x, y):
    out = model(x)
    y_hat = out.argmax(-1)
    corrects = (y_hat == y).cast(slope.float32)
    accuracy = corrects.mean()
    return accuracy


class Net(nn.Module):
    def __init__(self, in_dim, hid_dim, out_dim, act=nn.ReLU()):
        self.flatten_fn = nn.Fn(lambda x: x.reshape(shape=(x.shape[0], math.prod(x.shape[1:]))))
        self.linear1 = nn.Linear(in_dim, hid_dim)
        self.linear2 = nn.Linear(hid_dim, out_dim)
        self.act = act

    def __call__(self, x):
        x = self.flatten_fn(x)
        x = self.linear1(x)
        x = self.act(x)
        x = self.linear2(x)
        return x


if __name__ == "__main__":
    num_epochs = 3
    batch_size = 200  # TODO: must be multiple of dataset
    train_images, train_labels, test_images, test_labels = get_mnist()
    num_train = train_images.shape[0]
    num_complete_batches, leftover = divmod(num_train, batch_size)
    num_batches = num_complete_batches + bool(leftover)
    log_interval = 10
    model = Net(in_dim=784, hid_dim=100, out_dim=10, act=nn.ReLU())
    optimizer = nn.SGD(model, lr=1e-3, momentum=0.9, weight_decay=1e-4)
    # optimizer = nn.Adam(model, lr=1e-3)

    def data_stream():
        rng = np.random.RandomState(0)
        while True:
            perm = rng.permutation(num_train)
            for i in range(num_batches):
                batch_idx = perm[i * batch_size : (i + 1) * batch_size]
                x = slope.tensor(train_images[batch_idx])
                y_onehot = slope.tensor(train_labels[batch_idx]).one_hot(10).cast(slope.float32)
                yield x, y_onehot

    x_test, y_test = slope.tensor(test_images), slope.tensor(test_labels).cast(slope.int32)

    batches = data_stream()
    itercount = itertools.count()

    print("\nStarting training...")
    for epoch in range(num_epochs):
        start_time = time.time()
        for i in (pbar := tqdm(range(num_batches))):
            batch = next(batches)
            loss, model, optimizer = train_step(model, batch, optimizer)
            pbar.set_description(f"Train epoch: {epoch}, batch: {i}/{num_batches}, loss: {loss.numpy():.2f}")
            # if i % 10 == 0: print(f"Train epoch: {epoch}, batch: {i}/{num_batches}, loss: {loss.numpy():.2f}")
        epoch_time = time.time() - start_time

        test_acc = test_all(model, x_test, y_test)
        print(f"Epoch {epoch} in {epoch_time:0.2f} sec")
        print(f"Test set accuracy {test_acc.numpy():0.2f}")



# A set of code samples showing different usage of the ONNX Runtime Python API
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import numpy as np
import torch
import onnxruntime
import onnx
# from onnx import parser, printer

MODEL_FILE = '.model.onnx'
DEVICE_NAME = 'cuda' if torch.cuda.is_available() else 'cpu'
DEVICE_INDEX = 0     # Replace this with the index of the device you want to run on
DEVICE=f'{DEVICE_NAME}:{DEVICE_INDEX}'
# SIZE = ()
SIZE = (1, 3, 32, 32)

# A simple model to calculate addition of two tensors
def model():
    class Model(torch.nn.Module):
        def __init__(self):
            super(Model, self).__init__()

        def forward(self, x, y):
            # return x.split(2)
            # return x + y + torch.full(SIZE, fill_value=1, dtype=torch.int32)
            # z = torch.ones(1, dtype=torch.float32)#.expand(*SIZE)
            # return x + y + z
            # out = (x+y).split(2,2)
            return ~(x.bool()) & y.bool() & torch.full(SIZE, True)
            return out
            # return x + y + torch.rand_like(y, dtype=torch.flot32)
    return Model()

# Create an instance of the model and export it to ONNX graph format, with dynamic size for the data
def create_model(type: torch.dtype = torch.float32):
    sample_x = torch.ones(*SIZE, dtype=type)
    sample_y = torch.zeros(*SIZE, dtype=type)

    torch.onnx.export(model(), (sample_x, sample_y), MODEL_FILE, input_names=["x", "y"], output_names=["z"],
                               dynamic_dim={"x": {0 : "array_length_x"}, "y": {0: "array_length_y"}})
 
# Create an ONNX Runtime session with the provided model
def create_session(model: str) -> onnxruntime.InferenceSession:
    providers = ['CPUExecutionProvider']
    if torch.cuda.is_available():
        providers.insert(0, 'CUDAExecutionProvider')
    ret = onnxruntime.InferenceSession(model, providers=providers)
    return ret

# Run the model on CPU consuming and producing numpy arrays 
def run(x: np.array, y: np.array) -> np.array:
    session = create_session(MODEL_FILE)

    z = session.run(["z"], {"x": x, "y": y})
    
    return z[0]   

# Run the model on device consuming and producing ORTValues
def run_with_data_on_device(x: np.array, y: np.array) -> onnxruntime.OrtValue:
    session = create_session(MODEL_FILE)

    x_ortvalue = onnxruntime.OrtValue.ortvalue_from_numpy(x, DEVICE_NAME, DEVICE_INDEX)
    y_ortvalue = onnxruntime.OrtValue.ortvalue_from_numpy(y, DEVICE_NAME, DEVICE_INDEX)

    io_binding = session.io_binding()
    io_binding.bind_input(name='x', device_type=x_ortvalue.device_name(), device_id=0, element_type=x.dtype, shape=x_ortvalue.shape(), buffer_ptr=x_ortvalue.data_ptr())
    io_binding.bind_input(name='y', device_type=y_ortvalue.device_name(), device_id=0, element_type=y.dtype, shape=y_ortvalue.shape(), buffer_ptr=y_ortvalue.data_ptr())
    io_binding.bind_output(name='z', device_type=DEVICE_NAME, device_id=DEVICE_INDEX, element_type=x.dtype, shape=x_ortvalue.shape())
    session.run_with_iobinding(io_binding)

    z = io_binding.get_outputs()

    return z[0]

# Run the model on device consuming and producing native PyTorch tensors
def run_with_torch_tensors_on_device(x: torch.Tensor, y: torch.Tensor, np_type: np.dtype = np.float32, torch_type: torch.dtype = torch.float32) -> torch.Tensor:
    session = create_session(MODEL_FILE)

    binding = session.io_binding()

    x_tensor = x.contiguous()
    y_tensor = y.contiguous()

    binding.bind_input(
        name='x',
        device_type=DEVICE_NAME,
        device_id=DEVICE_INDEX,
        element_type=np_type,
        shape=tuple(x_tensor.shape),
        buffer_ptr=x_tensor.data_ptr(),
        )

    binding.bind_input(
        name='y',
        device_type=DEVICE_NAME,
        device_id=DEVICE_INDEX,
        element_type=np_type,
        shape=tuple(y_tensor.shape),
        buffer_ptr=y_tensor.data_ptr(),
        )

    ## Allocate the PyTorch tensor for the model output
    z_tensor = torch.empty(x_tensor.shape, dtype=torch_type, device=DEVICE).contiguous()
    binding.bind_output(
        name='z',
        device_type=DEVICE_NAME,
        device_id=DEVICE_INDEX,
        element_type=np_type,
        shape=tuple(z_tensor.shape),
        buffer_ptr=z_tensor.data_ptr(),
    )

    session.run_with_iobinding(binding)

    return z_tensor

def main():
    create_model()
    model = onnx.load(MODEL_FILE)
    # print(model)
    text = onnx.printer.to_text(model)
    # print(f"{len(text)=}")
    print(text)
    model_hat = onnx.parser.parse_model(text.replace('/', ''))
    # print(model_hat)
    # ones = np.ones(SIZE, dtype=bool)
    ones = np.ones(SIZE, dtype=np.float32)
    out = run(x=ones,y=ones)
    for o in out:print(o.shape)
    # [array([5., 7., 9.], dtype=float32)]

    # print(run_with_data_on_device(x=np.float32([1.0, 2.0, 3.0, 4.0, 5.0]), y=np.float32([1.0, 2.0, 3.0, 4.0, 5.0])).numpy())
    # # [ 2.  4.  6.  8. 10.]

    # print(run_with_torch_tensors_on_device(torch.rand(5).to(DEVICE), torch.rand(5).to(DEVICE)))
    # # tensor([0.7023, 1.3127, 1.7289, 0.3982, 0.8386])

    # create_model(torch.int64)
 
    # print(run_with_torch_tensors_on_device(torch.ones(5, dtype=torch.int64).to(DEVICE), torch.zeros(5, dtype=torch.int64).to(DEVICE), np_type=np.int64, torch_type=torch.int64))
    # # tensor([1, 1, 1, 1, 1])


if __name__ == "__main__":
    main()   